import datetime
import logging
from operator import itemgetter
import itertools
import collections
from pathlib import Path
from pdb import set_trace

import matplotlib as mpl
import netCDF4 as netcdf
import numpy as np
from matplotlib import pyplot as plt
from matplotlib.ticker import ScalarFormatter

NETCDF_FORMAT = 'NETCDF4'  # choose one of 'NETCDF3_CLASSIC', 'NETCDF3_64BIT', 'NETCDF4_CLASSIC' and 'NETCDF4'

logger = logging.getLogger(__name__)


class BaseLidarMeasurement(object):
    """ 
    This class represents a general measurement object, independent of the input files.
    
    Each subclass should implement the following:
    * the _import_file method;
    * set the "extra_netcdf_parameters" variable to a dictionary that includes the appropriate parameters;
    
    You can override the set_PT method to define a custom procedure to get ground temperature and pressure.   
    
    The class assumes that the input files are consecutive, i.e. there are no measurements gaps.
    """
    extra_netcdf_parameters = None

    def __init__(self, file_list=None):
        """
        This is run when creating a new object.
        
        Parameters
        ----------
        file_list : list or str
           A list of the full paths to the input file(s).
        """
        self.info = {}
        self.dimensions = {}
        self.variables = {}
        self.channels = collections.OrderedDict()
        self.attributes = {}
        self.files = []
        self.dark_measurement = None

        if file_list:
            self._import_files(file_list)

    def _import_files(self, file_list):
        """
        Imports a list of files, and updates the object parameters.
        
        Parameters
        ----------
        file_list : list
           A list of the full paths to the input file. 
        """
        for f in file_list:
            self._import_file(f)
        self.update()

    def _import_file(self, filename):
        """
        Reads a single lidar file.
        
        This method should be overwritten at all subclasses.
        
        Parameters
        ----------
        filename : str
           Path to the lidar file.
        """
        raise NotImplementedError('Importing files should be defined in the instrument-specific subclass.')

    def update(self):
        """
        Update the info dictionary, variables, and dimensions of the measurement object
        based on the information found in the channels objects.
        
        Notes
        -----
        Reading of the scan_angles parameter is not implemented.
        """
        # Initialize
        start_time = []
        stop_time = []
        points = []
        all_time_scales = []
        channel_names = []

        # Get the information from all the channels
        for channel_name, channel in self.channels.items():
            channel.update()
            start_time.append(channel.start_time)
            stop_time.append(channel.stop_time)
            points.append(channel.points)
            all_time_scales.append(channel.time)
            channel_names.append(channel_name)

        # Find the unique time scales used in the channels
        time_scales = set(all_time_scales)

        # Update the info dictionary
        self.info['start_time'] = min(start_time)
        self.info['stop_time'] = max(stop_time)
        self.info['duration'] = self.info['stop_time'] - self.info['start_time']

        # Update the dimensions dictionary
        self.dimensions['points'] = max(points)
        self.dimensions['channels'] = len(self.channels)
        # self.dimensions['scan angles'] = 1
        self.dimensions['nb_of_time_scales'] = len(time_scales)

        # Make a dictionaries to match time scales and channels
        channel_timescales = {}
        timescale_channels = dict((ts, []) for ts in time_scales)
        for (channel_name, current_time_scale) in zip(channel_names, all_time_scales):
            for (ts, n) in zip(time_scales, list(range(len(time_scales)))):
                if current_time_scale == ts:
                    channel_timescales[channel_name] = n
                    timescale_channels[ts].append(channel_name)

        self.variables['id_timescale'] = channel_timescales

        # Update the variables dictionary
        # Write time scales in seconds
        raw_data_start_time = []
        raw_data_stop_time = []

        for current_time_scale in list(time_scales):
            raw_start_time = np.array(current_time_scale) - min(start_time)  # Time since start_time
            raw_start_in_seconds = np.array([t.seconds for t in raw_start_time])  # Convert in seconds
            raw_data_start_time.append(raw_start_in_seconds)  # And add to the list

            channel_name = timescale_channels[current_time_scale][0]
            channel = self.channels[channel_name]

            # TODO: Define stop time for each measurement based on real data
            raw_stop_in_seconds = raw_start_in_seconds + channel.get_duration()
            raw_data_stop_time.append(raw_stop_in_seconds)

        self.variables['Raw_Data_Start_Time'] = raw_data_start_time
        self.variables['Raw_Data_Stop_Time'] = raw_data_stop_time

    def subset_by_channels(self, channel_subset):
        """ 
        Create a measurement object containing only a subset of  channels. 
        
        Parameters
        ----------
        channel_subset : list
           A list of channel names (str) to be included in the new measurement object.
        
        Returns
        -------
        m : BaseLidarMeasurements object
           A new measurements object
        """

        m = self.__class__()  # Create an object of the same type as this one.
        m.channels = dict([(channel, self.channels[channel]) for channel
                           in channel_subset])
        m.files = self.files

        m.update()

        # Dark measurements should also be subsetted.
        if self.dark_measurement is not None:
            dark_subset = self.dark_measurement.subset_by_channels(channel_subset)
            m.dark_measurement = dark_subset

        return m

    def subset_by_scc_channels(self):
        """
        Subset the measurement based on the channels provided in the 
        extra_netecdf_parameter file.
        
        Returns
        -------
        m : BaseLidarMeasurements object
           A new measurements object
        """
        if self.extra_netcdf_parameters is None:
            raise RuntimeError("Extra netCDF parameters not defined, cannot subset measurement.")

        scc_channels = list(self.extra_netcdf_parameters.channel_parameters.keys())
        common_channels = list(set(scc_channels).intersection(list(self.channels.keys())))

        if not common_channels:
            logger.debug("Config channels: %s." % ','.join(scc_channels))
            logger.debug("Measurement channels: %s." % ','.join(list(self.channels.keys())))
            raise ValueError('No common channels between measurement and configuration files.')

        return self.subset_by_channels(common_channels)

    def subset_by_time(self, start_time, stop_time):
        """
        Subset the measurement for a specific time interval
        
        Parameters
        ----------
        start_time : datetime 
           The starting datetime to subset.
        stop_time : datetime
           The stopping datetime to subset.

        Returns
        -------
        m : BaseLidarMeasurements object
           A new measurements object
        """

        if start_time > stop_time:
            raise ValueError('Stop time should be after start time')

        if (start_time < self.info['start_time']) or (stop_time > self.info['stop_time']):
            raise ValueError('The time interval specified is not part of the measurement')

        m = self.__class__()  # Create an object of the same type as this one.
        for channel_name, channel in self.channels.items():
            m.channels[channel_name] = channel.subset_by_time(start_time, stop_time)

        m.update()

        # Transfer dark measurement to the new object. They don't need time subsetting.
        m.dark_measurement = self.dark_measurement
        return m

    def subset_by_bins(self, b_min=0, b_max=None):
        """
        Remove some height bins from the measurement data. 
        
        This could be needed to remove acquisition artifacts at 
        the first or last bins of the profiles.
        
        Parameters
        ----------
        b_min : int
          The fist valid data bin
        b_max : int or None
          The last valid data bin. If equal to None, all bins are used.
          
        Returns
        -------
        m : BaseLidarMeasurements object
           A new measurements object
        """
        m = self.__class__()  # Create an object of the same type as this one.

        for channel_name, channel in self.channels.items():
            m.channels[channel_name] = channel.subset_by_bins(b_min, b_max)

        m.update()

        # Dark measurements should also be subseted.
        if self.dark_measurement is not None:
            dark_subset = self.dark_measurement.subset_by_bins(b_min, b_max)
            m.dark_measurement = dark_subset

        return m

    def rename_channels(self, prefix="", suffix=""):
        """ Add a prefix and a suffix to all channel name.
        
        This is uses when processing Delta90 depolarization calibration measurements.
        
        Parameters
        ----------
        prefix : str
          The prefix to add to channel names. 
        suffix : str
          The suffix to add to channel names.
        """
        channel_names = list(self.channels.keys())

        for channel_name in channel_names:
            new_name = prefix + channel_name + suffix
            self.channels[new_name] = self.channels.pop(channel_name)

    def set_PT(self):
        """ 
        Sets the pressure and temperature at station level at the info dictionary .
        
        In this method, default values are used. It can be overwritten by subclasses
        to define more appropriate values for each system.
        """
        self.info['Temperature'] = 15.0  # Temperature in degC
        self.info['Pressure'] = 1013.15  # Pressure in hPa

    def subtract_dark(self):
        """
        Subtract dark measurements from the raw lidar signals. 

        This method is here just for testing.
        
        Notes
        -----
        This method should not be called if processing the data with the SCC. The SCC
        performs this operations anyway. 
        """
        if not self.dark_measurement:
            raise IOError('No dark measurements have been imported yet.')

        for (channel_name, dark_channel) in self.dark_measurement.channels.items():
            dark_profile = dark_channel.average_profile()
            channel = self.channels[channel_name]

            for measurement_time, data in channel.data.items():
                channel.data[measurement_time] = data - dark_profile

            channel.update()

    def set_measurement_id(self, measurement_id):
        """
        Sets the measurement id for the SCC file.

        Parameters
        ----------
        measurement_id: str           
        """

        self.info['Measurement_ID'] = measurement_id

    def save_as_SCC_netcdf(self, filename=None, output_dir=None):
        """Saves the measurement in the netCDF format as required by the SCC.
        
        If no filename is provided <measurement_id>.nc will be used. 
        
        Parameters
        ----------
        filename : str
           Output file name. If None, <measurement_id>.nc will be used. 
        """
        parameters = self.extra_netcdf_parameters

        #Management output path
        if output_dir is None:
            output_dir = Path.cwd()

        # Guess measurement ID if none is provided
        if 'Measurement_ID' not in self.info:
            self.set_measurement_id()

        # Check if temperature and pressure are defined
        for attribute in ['Temperature', 'Pressure']:
            stored_value = self.info.get(attribute, None)
            if stored_value is None:
                try:
                    self.set_PT()
                except:
                    raise ValueError('No value specified for %s' % attribute)

        if not filename:
            filename = "%s.nc" % self.info['Measurement_ID']

        filepath = output_dir / filename

        self.scc_filename = filename

        dimensions = {'points': 1,
                      'channels': 1,
                      'time': None,
                      'nb_of_time_scales': 1,
                      'scan_angles': 1, }  # Mandatory dimensions. Time bck not implemented

        global_attributes = {'Measurement_ID': None,
                             'RawData_Start_Date': None,
                             'RawData_Start_Time_UT': None,
                             'RawData_Stop_Time_UT': None,
                             'RawBck_Start_Date': None,
                             'RawBck_Start_Time_UT': None,
                             'RawBck_Stop_Time_UT': None,
                             'Sounding_File_Name': None,
                             'LR_File_Name': None,
                             'Overlap_File_Name': None,
                             'Location': None,
                             'System': None,
                             'Latitude_degrees_north': None,
                             'Longitude_degrees_east': None,
                             'Altitude_meter_asl': None,}

        channel_names = list(self.channels.keys())

        input_values = dict(self.dimensions, **self.variables)

        # Add some mandatory global attributes
        input_values['Measurement_ID'] = self.info['Measurement_ID']
        input_values['RawData_Start_Date'] = self.info['start_time'].strftime('%Y%m%d')
        input_values['RawData_Start_Time_UT'] = self.info['start_time'].strftime('%H%M%S')
        input_values['RawData_Stop_Time_UT'] = self.info['stop_time'].strftime('%H%M%S')

        # Add any global attributes provided by the subclass
        for attribute in self._get_custom_global_attributes():
            input_values[attribute["name"]] = attribute["value"]

        # Override global attributes, if provided in the settings file.
        for attribute_name in ['System', 'Latitude_degrees_north', 'Longitude_degrees_east', 'Altitude_meter_asl',
                               'Overlap_File_Name', 'LR_File_Name', 'Sounding_File_Name']:
            if attribute_name in parameters.general_parameters.keys():
                if attribute_name in input_values:
                    logger.info("Overriding {0} attribute, using the value provided in the parameter file.".format(
                        attribute_name))
                input_values[attribute_name] = parameters.general_parameters[attribute_name]
        # Open a netCDF file. The format is specified in the beginning of this module.
        with netcdf.Dataset(filepath, 'w', format=NETCDF_FORMAT) as f:

            # Create the dimensions in the file
            for (d, v) in dimensions.items():
                v = input_values.pop(d, v)
                f.createDimension(d, v)

            # Create global attributes
            for (attrib, value) in global_attributes.items():
                val = input_values.pop(attrib, value)
                if val:
                    setattr(f, attrib, val)
            
            # Variables
            # Write either channel_id or string_channel_id in the file
            first_channel_keys = list(parameters.channel_parameters.items())[0][1].keys()
            if "channel_ID" in first_channel_keys:
                channel_var = 'channel_ID'
                variable_type = 'i'
            elif "channel_string_ID" in first_channel_keys:
                channel_var = 'channel_string_ID'
                variable_type = str
            else:
                raise ValueError('Channel parameters should define either "chanel_id" or "channel_string_ID".')

            temp_v = f.createVariable(channel_var, variable_type, ('channels',))
            for n, channel in enumerate(channel_names):
                temp_v[n] = parameters.channel_parameters[channel][channel_var]

            # Write the custom subclass variables:
            for variable in self._get_custom_variables(channel_names):
                logger.debug("Saving variable {0}".format(variable['name']))
                temp_v = f.createVariable(variable["name"], variable["type"], variable["dimensions"])
                temp_v[:] = variable["values"]

            # Write the values of fixed channel parameters:            
            channel_variable_specs = self._get_scc_channel_variables()
            if 'First_Signal_Rangebin' not in channel_variable_specs.keys():
                channel_variable_specs['First_Signal_Rangebin'] = (('channels',), 'i')
            if 'Background_Mode' not in channel_variable_specs.keys():
                channel_variable_specs['Background_Mode'] = (('channels',), 'i')                
            provided_variable_names = self._get_provided_variable_names()

            for variable_name in provided_variable_names:
                # Check if the variable already defined (e.g. from values in Licel files).
                if variable_name in f.variables.keys():
                    logger.warning(
                        "Provided values of \"{0}\" were ignored because they were read from other source.".format(
                            variable_name))
                    continue

                if variable_name not in channel_variable_specs.keys():
                    logger.warning("Provided variable {0} is not parts of the specs: {1}".format(variable_name, list(channel_variable_specs.keys())))
                    continue

                temp_v = f.createVariable(variable_name,
                                          channel_variable_specs[variable_name][1],
                                          channel_variable_specs[variable_name][0])

                for n, channel_name in enumerate(channel_names):
                    try:
                        temp_v[n] = parameters.channel_parameters[channel_name][variable_name]
                    except KeyError:  # The parameter was not provided for this channel so we mask the value.
                        pass  # Default value should be a missing value  # TODO: Check this.

            # Write the id_timescale values
            temp_id_timescale = f.createVariable('id_timescale', 'i', ('channels',))
            for (channel, n) in zip(channel_names, list(range(len(channel_names)))):
                temp_id_timescale[n] = self.variables['id_timescale'][channel]

            # Laser pointing angle
            temp_v = f.createVariable('Laser_Pointing_Angle', 'd', ('scan_angles',))
            temp_v[:] = parameters.general_parameters['Laser_Pointing_Angle']

            # Molecular calculation
            temp_v = f.createVariable('Molecular_Calc', 'i')
            temp_v[:] = parameters.general_parameters['Molecular_Calc']

            # Laser pointing angles of profiles
            temp_v = f.createVariable('Laser_Pointing_Angle_of_Profiles', 'i', ('time', 'nb_of_time_scales'))
            for (time_scale, n) in zip(self.variables['Raw_Data_Start_Time'],
                                       list(range(len(self.variables['Raw_Data_Start_Time'])))):
                temp_v[:len(time_scale), n] = 0  # The lidar has only one laser pointing angle

            # Raw data start/stop time
            temp_raw_start = f.createVariable('Raw_Data_Start_Time', 'i', ('time', 'nb_of_time_scales'))
            temp_raw_stop = f.createVariable('Raw_Data_Stop_Time', 'i', ('time', 'nb_of_time_scales'))
            for (start_time, stop_time, n) in zip(self.variables['Raw_Data_Start_Time'],
                                                  self.variables['Raw_Data_Stop_Time'],
                                                  list(range(len(self.variables['Raw_Data_Start_Time'])))):
                temp_raw_start[:len(start_time), n] = start_time
                temp_raw_stop[:len(stop_time), n] = stop_time

            # Laser shots
            if "Laser_Shots" in provided_variable_names:
                if "Laser_Shots" in f.variables.keys():
                    logger.warning("Provided values of \"Laser_Shots\" were ignored because they were read from other source.")
                else:
                    temp_v = f.createVariable('Laser_Shots', 'i', ('time', 'channels'))
                    for (channel, n) in zip(channel_names, list(range(len(channel_names)))):
                        time_length = len(self.variables['Raw_Data_Start_Time'][self.variables['id_timescale'][channel]])
                        # Array slicing stopped working as usual ex. temp_v[:10] = 100 does not work. ??? np.ones was added.
                        temp_v[:time_length, n] = np.ones(time_length) * parameters.channel_parameters[channel][
                            'Laser_Shots']
            else:
                if "Laser_Shots" not in f.variables.keys():
                    logger.error("Mandatory variable \"Laser_Shots\" was not found in the settings or input files.")
                else:
                    pass  # Laser shots already in variables, so all good.

            # Raw lidar data
            temp_v = f.createVariable('Raw_Lidar_Data', 'd', ('time', 'channels', 'points'), zlib=True)
            for (channel, n) in zip(channel_names, list(range(len(channel_names)))):
                c = self.channels[channel]
                temp_v[:len(c.time), n, :c.points] = c.matrix

            self.add_dark_measurements_to_netcdf(f, channel_names)

            # Pressure at lidar station
            temp_v = f.createVariable('Pressure_at_Lidar_Station', 'd')
            temp_v[:] = self.info['Pressure']

            # Temperature at lidar station
            temp_v = f.createVariable('Temperature_at_Lidar_Station', 'd')
            temp_v[:] = self.info['Temperature']

            self.save_netcdf_extra(f)

    def _get_scc_channel_variables(self):        
        channel_variables = \
            {'Laser_Repetition_Rate': (('channels',), 'i'),
             'Scattering_Mechanism': (('channels',), 'i'),
             'Signal_Type': (('channels',), 'i'),
             'Emitted_Wavelength': (('channels',), 'd'),
             'Detected_Wavelength': (('channels',), 'd'),
             'Raw_Data_Range_Resolution': (('channels',), 'd'),
             'Background_Mode': (('channels',), 'i'),
             'Background_Low': (('channels',), 'd'),
             'Background_High': (('channels',), 'd'),
             'Dead_Time_Corr_Type': (('channels',), 'i'),
             'Dead_Time': (('channels',), 'd'),
             'Acquisition_Mode': (('channels',), 'i'),
             'Trigger_Delay': (('channels',), 'd'),
             'LR_Input': (('channels',), 'i'),
             'DAQ_Range': (('channels',), 'd'),
             'First_Signal_Rangebin': (('channels',), 'i'),
             }        
        return channel_variables

    def _get_provided_variable_names(self):
        """ Return a list of """
        # When looking for channel parameters, ignore the following parameter names:
        ignore = {'channel_ID', 'channel_string_ID', 'Depolarization_Factor', 'Laser_Shots'}  # Set

        channels = list(self.channels.keys())
        channel_parameters = self.extra_netcdf_parameters.channel_parameters

        # Get all the provided parameters (both mandatory and optional):
        all_provided_variables = [list(channel_parameters[channel].keys()) for channel in channels]
        provided_variables = set(itertools.chain.from_iterable(all_provided_variables))

        # Discard certain parameter names:
        provided_variables -= ignore

        return provided_variables

    def add_dark_measurements_to_netcdf(self, f, channels):
        """
        Adds dark measurement variables and properties to an open netCDF file.
        
        Parameters
        ----------
        f : netcdf Dataset
           A netCDF Dataset, open for writing.
        channels : list
           A list of channels names to consider when adding dark measurements.
        """
        # Get dark measurements. If it is not given in self.dark_measurement
        # try to get it using the get_dark_measurements method. If none is found
        # return without adding something.
        if self.dark_measurement is None:
            self.dark_measurement = self.get_dark_measurements()

        if self.dark_measurement is None:
            return

        dark_measurement = self.dark_measurement

        # Calculate the length of the time_bck dimensions
        number_of_profiles = [len(c.time) for c in dark_measurement.channels.values()]
        max_number_of_profiles = np.max(number_of_profiles)

        # Create the dimension
        f.createDimension('time_bck', max_number_of_profiles)

        # Save the dark measurement data
        temp_v = f.createVariable('Background_Profile', 'd', ('time_bck', 'channels', 'points'), zlib=True)
        for (channel, n) in zip(channels, list(range(len(channels)))):
            c = dark_measurement.channels[channel]
            temp_v[:len(c.time), n, :c.points] = c.matrix

        # Dark profile start/stop time
        temp_raw_start = f.createVariable('Raw_Bck_Start_Time', 'i', ('time_bck', 'nb_of_time_scales'))
        temp_raw_stop = f.createVariable('Raw_Bck_Stop_Time', 'i', ('time_bck', 'nb_of_time_scales'))
        for (start_time, stop_time, n) in zip(dark_measurement.variables['Raw_Data_Start_Time'],
                                              dark_measurement.variables['Raw_Data_Stop_Time'],
                                              list(range(len(dark_measurement.variables['Raw_Data_Start_Time'])))):
            temp_raw_start[:len(start_time), n] = start_time
            temp_raw_stop[:len(stop_time), n] = stop_time

        # Dark measurement start/stop time
        f.RawBck_Start_Date = dark_measurement.info['start_time'].strftime('%Y%m%d')
        f.RawBck_Start_Time_UT = dark_measurement.info['start_time'].strftime('%H%M%S')
        f.RawBck_Stop_Time_UT = dark_measurement.info['stop_time'].strftime('%H%M%S')

    def save_netcdf_extra(self, f):
        """ Save extra netCDF parameters to an open netCDF file. 
        
        If required, this method should be overwritten by subclasses of BaseLidarMeasurement.
        """
        pass

    def plot(self):
        for channel in self.channels:
            self.channels[channel].plot(show_plot=False)
        plt.show()

    def get_dark_measurements(self):
        return None

    def _get_custom_global_attributes(self):
        """
        Abstract method to provide NetCDF global attributes that should be included
        in the final NetCDF file.
        
        If required, this method should be implemented by subclasses of BaseLidarMeasurement.
        """
        return []

    def _get_custom_variables(self, channel_names=None):
        """
        Abstract method to provide custom NetCDF variables that should be included in the final NetCDF file.
        
        If required, this method should be implemented by subclasses of BaseLidarMeasurement.
        """
        return []


class LidarChannel(object):
    """ 
    This class represents a general measurement channel, independent of the input files.

    The class assumes that the input files are consecutive, i.e. there are no measurements gaps.
    """

    def __init__(self, channel_parameters):
        """
        This is run when creating a new object.
        
        The input dictionary should contain 'name', 'binwidth', and 'data' keys. 
        
        Parameters
        ----------
        channel_parameters : dict
           A dict containing channel parameters.
        """
        # TODO: Change channel_parameters to explicit keyword arguments?

        c = 299792458.  # Speed of light
        self.wavelength = channel_parameters['name']
        self.name = str(self.wavelength)
        self.binwidth = float(channel_parameters['binwidth'])  # in microseconds
        self.data = {}
        self.resolution = self.binwidth * c / 2
        self.z = np.arange(
            len(channel_parameters['data'])) * self.resolution + self.resolution / 2.0  # Change: add half bin in the z
        self.points = len(channel_parameters['data'])
        self.rc = []

    def get_duration(self):
        """ Get an array with the duration of each profile in seconds.

        If the duration property already exists, returns this.
        If not, it tries to estimate it based on the time difference of the profiles.

        Returns
        -------
        duration : ndarray
           1D array containing profile durations.
        """

        current_value = getattr(self, 'duration', None)

        if current_value:
            return np.array(current_value)

        time_diff = np.diff(self.time)
        durations = [dt.seconds for dt in time_diff]
        # Assume the last two profiles have the same duration
        duration = np.array(durations)
        return duration

    def calculate_rc(self, idx_min=-2000, idx_max=-500, first_signal_bin=0):
        """ Calculate range corrected signal.
        
        The background signal is estimated as the mean signal between idx_min and idx_max. 
        
        The calculations is stored in the self.rc parameters. 
        
        Parameters
        ----------
        idx_min : int
           Minimum index to calculate background signal.
        idx_max : int
           Maximum index to calculate background signal.
        first_signal_bin : int
           First bin with useful signal. Positive integer.
        """
        background = np.mean(self.matrix[:, idx_min:idx_max], axis=1)
        background_corrected = (self.matrix.transpose() - background).transpose()
        background_corrected = np.roll(background_corrected, -first_signal_bin, axis=1)
        background_corrected[:, -first_signal_bin:] = 0
        self.rc = background_corrected * (self.z ** 2)

    def noise_mask(self, idx_min=-2000, idx_max=-500, threshold=1.):
        """ Calculate points that are probably noise.

        To calculate this, we find the max value of the background region. Then we reject all points that
        are less than `threshold` times this value.
        Parameters
        ----------
        idx_min : int
           Minimum index to calculate background signal.
        idx_max : int
           Maximum index to calculate background signal.
        threshold : float
           Threshold value.
        """
        background_max = np.max(self.matrix[:, idx_min:idx_max], axis=1)
        background_mean = np.mean(self.matrix[:, idx_min:idx_max], axis=1)
        mean_to_max = background_max - background_mean

        noise_limit = background_mean + mean_to_max * threshold

        mask = self.matrix <= noise_limit[:, np.newaxis]

        return mask

    def update(self):
        """
        Update the time parameters and data according to the raw input data. 
        """
        self.start_time = min(self.data.keys())
        self.stop_time = max(self.data.keys()) + datetime.timedelta(seconds=self.duration[-1])
        self.time = tuple(sorted(self.data.keys()))
        sorted_data = sorted(iter(self.data.items()), key=itemgetter(0))
        self.matrix = np.array(list(map(itemgetter(1), sorted_data)))

    def _nearest_datetime(self, input_time):
        """
        Find the profile datetime and index that is nearest to the given datetime.
        
        Parameters
        ----------
        input_time : datetime.datetime
           Input datetime object.

        Returns
        -------
        profile_datetime : datetime
           The datetime of the selected profile.
        profile_idx : int
           The index of the selected profile.
        """
        max_duration = np.max(self.duration)

        margin = datetime.timedelta(seconds=max_duration * 5)

        if ((input_time + margin) < self.start_time) | ((input_time - margin) > self.stop_time):
            logger.error("Requested date not covered in this file")
            raise ValueError("Requested date not covered in this file")
        dt = np.abs(np.array(self.time) - input_time)

        profile_idx = np.argmin(dt)
        profile_datetime = self.time[profile_idx]

        dt_min = dt[profile_idx]
        if dt_min > datetime.timedelta(seconds=max_duration):
            logger.warning("Nearest profile more than %s seconds away. dt = %s." % (max_duration, dt_min))

        return profile_datetime, profile_idx

    def subset_by_time(self, start_time, stop_time):
        """
        Subset the channel for a specific time interval.

        Parameters
        ----------
        start_time : datetime 
           The starting datetime to subset.
        stop_time : datetime
           The stopping datetime to subset.

        Returns
        -------
        c : LidarChannel object
           A new channel object
        """
        time_array = np.array(self.time)
        condition = (time_array >= start_time) & (time_array <= stop_time)

        subset_time = time_array[condition]
        subset_data = dict([(c_time, self.data[c_time]) for c_time in subset_time])

        # Create a list with the values needed by channel's __init__()
        parameter_values = {'name': self.wavelength,
                            'binwidth': self.binwidth,
                            'data': subset_data[subset_time[0]], }

        # We should use __class__ instead of class name, so that this works properly
        # with subclasses
        # Eg: c = self.__class__(parameters_values)
        # This does not currently work with Licel files though
        # TODO: Fix this!
        c = LidarChannel(parameter_values)
        c.data = subset_data
        c.update()
        return c

    def subset_by_bins(self, b_min=0, b_max=None):
        """
        Remove some height bins from the channel data. 
        
        This could be needed to remove acquisition artifacts at 
        the first or last bins of the profiles.
        
        Parameters
        ----------
        b_min : int
          The fist valid data bin
        b_max : int or None
          The last valid data bin. If equal to None, all bins are used.
          
        Returns
        -------
        m : LidarChannel object
           A new channel object
        """

        subset_data = {}

        for ctime, cdata in self.data.items():
            subset_data[ctime] = cdata[b_min:b_max]

        # Create a list with the values needed by channel's __init__()
        parameters_values = {'name': self.wavelength,
                             'binwidth': self.binwidth,
                             'data': subset_data[
                                 list(subset_data.keys())[0]], }  # We just need any array with the correct length

        c = LidarChannel(parameters_values)
        c.data = subset_data
        c.update()
        return c

    def get_profile(self, input_datetime, range_corrected=True):
        """ Returns a single profile, that is nearest to the input datetime.
        
        Parameters
        ----------
        input_datetime : datetime
           The required datetime of the profile
        range_corrected : bool
           If True, the range corrected profile is returned. Else, normal signal is returned.
        
        Returns
        -------
        profile : ndarray
           The profile nearest to input_datetime.
        t : datetime
           The actual datetime corresponding to the profile.
        """
        t, idx = self._nearest_datetime(input_datetime)
        if range_corrected:
            data = self.rc
        else:
            data = self.matrix

        profile = data[idx, :][0]

        return profile, t

    def get_slice(self, start_time, stop_time, range_corrected=True):
        """ Returns a slice of data, between the provided start and stop time.

        Parameters
        ----------
        start_time : datetime
           The starting time of the slice
        stop_time : datetime
           The stoping time of the slice
        range_corrected : bool
           If True, the range corrected profile is returned. Else, normal signal is returned.

        Returns
        -------
        slice : ndarray
           The slice of profiles.
        slice_datetimes : ndarray of datetime
           The actual datetimes corresponding to the slice.
        """
        if range_corrected:
            data = self.rc
        else:
            data = self.matrix

        time_array = np.array(self.time)
        start_time = self._nearest_datetime(start_time)[0]
        stop_time = self._nearest_datetime(stop_time)[0]

        condition = (time_array >= start_time) & (time_array <= stop_time)

        slice = data[condition, :]
        slice_datetimes = time_array[condition]

        return slice, slice_datetimes

    def average_profile(self):
        """ Return the average profile (NOT range corrected) for all the duration of 
        the measurement. 
        
        Returns
        -------
        profile : ndarray
           The average profile for all data.
        """
        profile = np.mean(self.matrix, axis=0)
        return profile

    def plot(self, figsize=(8, 4), signal_type='rc', zoom=[0, 12000, 0, None], show_plot=True,
             cmap=plt.cm.jet, z0=None, title=None, vmin=0, vmax=1.3 * 10 ** 7):
        """
        Plot of the channel data.
        
        Parameters
        ----------
        figsize : tuple
           (width, height) of the output figure (inches)
        signal_type : str
           If 'rc', the range corrected signal is ploted. Else, the raw signals are used.
        zoom : list
           A four elemet list defined as [xmin, xmax, ymin, ymax]. Use ymin=0, ymax=-1 to plot full range.
        show_plot : bool
           If True, the show_plot command is run. 
        cmap : cmap
           An instance of a matplotlib colormap to use.
        z0 : float
           The ground-level altitude. If provided the plot shows altitude above sea level.
        title : str
           Optional title for the plot.
        vmin : float
           Minimum value for the color scale.
        vmax : float
           Maximum value for the color scale.
        """
        fig = plt.figure(figsize=figsize)
        ax1 = fig.add_subplot(111)

        self.draw_plot_new(ax1, cmap=cmap, signal_type=signal_type, zoom=zoom, z0=z0, vmin=vmin, vmax=vmax)

        if title:
            ax1.set_title(title)
        else:
            ax1.set_title("%s signal - %s" % (signal_type.upper(), self.name))

        if show_plot:
            plt.show()

    def draw_plot(self, ax1, cmap=plt.cm.jet, signal_type='rc',
                  zoom=[0, 12000, 0, None], z0=None,
                  add_colorbar=True, cmap_label='a.u.', cb_format=None,
                  vmin=0, vmax=1.3 * 10 ** 7):
        """
        Draw channel data on the given axis.
        
        Parameters
        ----------
        ax1 : axis object
           The axis object to draw.
        cmap : cmap
           An instance of a matplotlib colormap to use.
        signal_type : str
           If 'rc', the range corrected signal is ploted. Else, the raw signals are used.
        zoom : list
           A four elemet list defined as [xmin, xmax, ymin, ymax]. Use ymin=0, ymax=-1 to plot full range.
        z0 : float
           The ground-level altitude. If provided the plot shows altitude above sea level.
        add_colorbar : bool
           If True, a colorbar will be added to the plot.
        cmap_label : str
           Label for the colorbar. Ignored if add_colorbar is False.
        cb_format : str
           Colorbar tick format string.
        vmin : float
           Minimum value for the color scale.
        vmax : float
           Maximum value for the color scale.
        """
        if signal_type == 'rc':
            if len(self.rc) == 0:
                self.calculate_rc()
            data = self.rc
        else:
            data = self.matrix

        hmax_idx = self._index_at_height(zoom[1])

        # If z0 is given, then the plot is a.s.l.
        if z0:
            ax1.set_ylabel('Altitude a.s.l. [km]')
        else:
            ax1.set_ylabel('Altitude a.g.l. [km]')
            z0 = 0

        ax1.set_xlabel('Time UTC [hh:mm]')
        # y axis in km, xaxis /2 to make 30s measurements in minutes. Only for 1064
        # dateFormatter = mpl.dates.DateFormatter('%H.%M')
        # hourlocator = mpl.dates.HourLocator()

        # dayFormatter = mpl.dates.DateFormatter('\n\n%d/%m')
        # daylocator = mpl.dates.DayLocator()
        hourFormatter = mpl.dates.DateFormatter('%H:%M')
        hourlocator = mpl.dates.AutoDateLocator(minticks=3, maxticks=8, interval_multiples=True)

        # ax1.axes.xaxis.set_major_formatter(dayFormatter)
        # ax1.axes.xaxis.set_major_locator(daylocator)
        ax1.axes.xaxis.set_major_formatter(hourFormatter)
        ax1.axes.xaxis.set_major_locator(hourlocator)

        ts1 = mpl.dates.date2num(self.start_time)
        ts2 = mpl.dates.date2num(self.stop_time)

        im1 = ax1.imshow(data.transpose()[zoom[0]:hmax_idx, zoom[2]:zoom[3]],
                         aspect='auto',
                         origin='lower',
                         cmap=cmap,
                         vmin=vmin,
                         # vmin = data[:,10:400].max() * 0.1,
                         vmax=vmax,
                         # vmax = data[:,10:400].max() * 0.9,
                         extent=[ts1, ts2, self.z[zoom[0]] / 1000.0 + z0 / 1000.,
                                 self.z[hmax_idx] / 1000.0 + z0 / 1000.],
                         )

        if add_colorbar:
            if cb_format:
                cb1 = plt.colorbar(im1, format=cb_format)
            else:
                cb1 = plt.colorbar(im1)
            cb1.ax.set_ylabel(cmap_label)

            # Make the ticks of the colorbar smaller, two points smaller than the default font size
            cb_font_size = mpl.rcParams['font.size'] - 2
            for ticklabels in cb1.ax.get_yticklabels():
                ticklabels.set_fontsize(cb_font_size)
            cb1.ax.yaxis.get_offset_text().set_fontsize(cb_font_size)

    def draw_plot_new(self, ax1, cmap=plt.cm.jet, signal_type='rc',
                      zoom=[0, 12000, 0, None], z0=None,
                      add_colorbar=True, cmap_label='a.u.',
                      cb_format=None, power_limits=(-2, 2),
                      date_labels=False,
                      vmin=0, vmax=1.3 * 10 ** 7):

        """ 
        Updated drawing routine, using pcolormesh. It is slower but provides more flexibility / precision
        in time plotting. The 2 drawing routines should be merged.
        
        Check draw_plot method for the meaning of the input arguments.
        """
        # TODO: Merge the two drawing routines.

        if signal_type == 'rc':
            if len(self.rc) == 0:
                self.calculate_rc()
            data = self.rc
        else:
            data = self.matrix

        self.draw_data(ax1, data, add_colorbar, cmap, cb_format, cmap_label, date_labels, power_limits, vmax, vmin,
                       z0, zoom)

    def draw_data(self, ax1, data, add_colorbar, cmap, cb_format, cmap_label, date_labels, power_limits, vmax, vmin,
                  z0, zoom):

        """
        This method plots generic lidar data in 2d Plot.

        Updated drawing routine, using pcolormesh. It is slower but provides more flexibility / precision
        in time plotting. The 2 drawing routines should be merged.


        """
        hmax_idx = self._index_at_height(zoom[1])
        hmin_idx = self._index_at_height(zoom[0])

        # If z0 is given, then the plot is a.s.l.
        if z0:
            ax1.set_ylabel('Altitude a.s.l. [km]')
        else:
            ax1.set_ylabel('Altitude a.g.l. [km]')
            z0 = 0
        ax1.set_xlabel('Time UTC [hh:mm]')

        # y axis in km, xaxis /2 to make 30s measurements in minutes. Only for 1064
        # dateFormatter = mpl.dates.DateFormatter('%H.%M')
        # hourlocator = mpl.dates.HourLocator()
        if date_labels:
            dayFormatter = mpl.dates.DateFormatter('%H:%M\n%d/%m/%Y')
            daylocator = mpl.dates.AutoDateLocator(minticks=3, maxticks=8, interval_multiples=True)
            ax1.axes.xaxis.set_major_formatter(dayFormatter)
            ax1.axes.xaxis.set_major_locator(daylocator)
        else:
            hourFormatter = mpl.dates.DateFormatter('%H:%M')
            hourlocator = mpl.dates.AutoDateLocator(minticks=3, maxticks=8, interval_multiples=True)
            ax1.axes.xaxis.set_major_formatter(hourFormatter)
            ax1.axes.xaxis.set_major_locator(hourlocator)

        # Get the values of the time axis
        dt = datetime.timedelta(seconds=self.duration[-1])

        time_cut = self.time[zoom[2]:zoom[3]]

        time_last = time_cut[-1] + dt  # The last element needed for pcolormesh
        time_all = time_cut + (time_last,)
        t_axis = mpl.dates.date2num(time_all)

        # Get the values of the z axis
        z_cut = self.z[hmin_idx:hmax_idx] - self.resolution / 2.
        z_last = z_cut[-1] + self.resolution
        z_axis = np.append(z_cut, z_last) / 1000. + z0 / 1000.  # Convert to km
        # Plot
        im1 = ax1.pcolormesh(t_axis, z_axis, data.T[hmin_idx:hmax_idx, zoom[2]:zoom[3]],
                             cmap=cmap,
                             vmin=vmin,
                             vmax=vmax,
                             )
        if add_colorbar:
            if cb_format:
                cb1 = plt.colorbar(im1, format=cb_format, extend='both')
            else:
                cb_formatter = ScalarFormatter()
                cb_formatter.set_powerlimits(power_limits)
                cb1 = plt.colorbar(im1, format=cb_formatter, extend='both')
            cb1.ax.set_ylabel(cmap_label)

            # Make the ticks of the colorbar smaller, two points smaller than the default font size
            cb_font_size = mpl.rcParams['font.size'] - 2
            for ticklabels in cb1.ax.get_yticklabels():
                ticklabels.set_fontsize(cb_font_size)
            cb1.ax.yaxis.get_offset_text().set_fontsize(cb_font_size)

    def _index_at_height(self, height):
        """
        Get the altitude index nearest to the specified height.
        
        Parameters
        ----------
        height : float
           Height (m)

        Returns
        -------
        idx : int
           Index corresponding to the provided height.
        """
        idx = np.array(np.abs(self.z - height).argmin())
        if idx.size > 1:
            idx = idx[0]
        return idx



# alh_daytime_an
# INSERT HERE THE SYSTEM PARAMETERS
general_parameters = {
    "System": "ALHAMBRA",
    "Laser_Pointing_Angle": 3,
    "Molecular_Calc": 0,  # Use US standard atmosphere
    "Latitude_degrees_north": 37.16,
    "Longitude_degrees_east": -3.60,
    "Altitude_meter_asl": 660.0,
    "Call sign": "gr",
}

# LINK YOUR LICEL CHANNELS TO SCC PARAMETERS. USE BT0, BC0 ETC AS NAMES (AS IN LICEL FILES).
channel_parameters = {
    "BT0": {
        "channel_ID": 2203,  # 1064fta
        "Background_Low": 40000.0,
        "Background_High": 50000.0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "DAQ_Range": 500.0,
    },
    "BT2": {
        "channel_ID": 2204,  # 532fta
        "Background_Low": 40000.0,
        "Background_High": 50000.0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "DAQ_Range": 100.0,
    },
    "BC2": {
        "channel_ID": 2232,  # 532ftp
        "Background_Low": 40000.0,
        "Background_High": 50000.0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
    },
    "BT10": {
        "channel_ID": 2123,  # 1064nta
        "Background_Low": 40000.0,
        "Background_High": 50000.0,
        "Laser_Shots": 600,
        "LR_Input": 1,
        "DAQ_Range": 500.0,
    },
    # "BT4": {
    #     "channel_ID": 2213,  # 355fpa
    #     "Background_Low": 40000.0,
    #     "Background_High": 50000.0,
    #     "Laser_Shots": 1200,
    #     "LR_Input": 1,
    #     "DAQ_Range": 100.0,
    # },
    # "BC4": {
    #     "channel_ID": 2233,  # 355fpp
    #     "Background_Low": 40000.0,
    #     "Background_High": 50000.0,
    #     "Laser_Shots": 1200,
    #     "LR_Input": 1,
    # },
    # "BT5": {
    #     "channel_ID": 2214,  # 355fsa
    #     "Background_Low": 40000.0,
    #     "Background_High": 50000.0,
    #     "Laser_Shots": 1200,
    #     "LR_Input": 1,
    #     "DAQ_Range": 100.0,
    # },
    # "BC5": {
    #     "channel_ID": 2234,  # 355fsp
    #     "Background_Low": 40000.0,
    #     "Background_High": 50000.0,
    #     "Laser_Shots": 1200,
    #     "LR_Input": 1,
    # },
}



from pathlib import Path
from typing import Any
import numpy as np
import xarray as xr

""" MODULE For General Radar Utilities
"""

def check_is_netcdf(path) -> Path:
    """Check if path is a netcdf file

    Args:

        - path (_type_): Radar file path.

    Raises:
        - FileNotFoundError: File does not exist.
        - ValueError: File is not a netcdf file.

    Returns:
        - Path: Path to netcdf file.
    """    
    
    if isinstance(path, str):
        path = Path(path)
    if not path.exists():
        raise FileNotFoundError(f"File {path} does not exist.")
    if path.suffix.lower() != ".nc":
        raise ValueError(f"File {path} is not a netcdf file.")
    return path

def enhance_rpgpy_dataset(
    path: Path, time_parser: str = "seconds since 2001-01-01 00:00:00 UTC"
) -> xr.Dataset:
    data = xr.open_dataset(path)
    if "Time" in data:
        data = data.rename({"Time": "time"})
    if "range_layers" in data:
        data["range"] = data["range_layers"]
    data.time.attrs["standard_name"] = "time"
    data.time.attrs["long_name"] = "time"
    data.time.attrs["units"] = time_parser
    data = xr.decode_cf(data)    
    return data


def ppi_to_cartessian(
    ranges: np.ndarray[np.float64] | xr.DataArray,
    azimuth: np.ndarray[np.float64] | xr.DataArray,
    elevation: np.ndarray[np.float64] | xr.DataArray,
) -> tuple[np.ndarray[np.float64], np.ndarray[np.float64]]:
    """Convert PPI coordinates to cartessian coordinates

    Args:
        ranges (np.ndarray[np.float64] | xr.DataArray): Range of the radar.
        azimuth (np.ndarray[np.float64] | xr.DataArray): Azimuth angle.
        elevation (np.ndarray[np.float64] | xr.DataArray): Elevation angle.

    Raises:
        ValueError: Elevation angle is not constant.

    Returns:
        tuple[np.ndarray[np.float64], np.ndarray[np.float64]]: x, y coordinates in cartessian plane [units as `ranges`].
    """    
    
    if isinstance(ranges, xr.DataArray):
        ranges = ranges.values
    if isinstance(azimuth, xr.DataArray):
        azimuth = azimuth.values
    if isinstance(elevation, xr.DataArray):
        elevation = elevation.values

    # Check elevation angle is constant
    if not np.allclose(elevation, elevation[0], atol=0.1):
        raise ValueError("Elevation angle is not constant.")
    elevation = elevation[0]
    rho = ranges * np.cos(np.deg2rad(elevation))
    x = rho[:, np.newaxis]  * np.sin(np.deg2rad(azimuth))
    y = rho[:, np.newaxis]  * np.cos(np.deg2rad(azimuth))
    return x, y

def rhi_to_cartessian(
    ranges: np.ndarray[np.float64] | xr.DataArray,    
    azimuth: np.ndarray[np.float64] | xr.DataArray,
    elevation: np.ndarray[np.float64] | xr.DataArray,
    ) -> tuple[np.ndarray[np.float64], np.ndarray[np.float64]]:
    """Convert RHI coordinates to cartessian coordinates

    Args:
        ranges (np.ndarray[np.float64] | xr.DataArray): Range of the radar.
        azimuth (np.ndarray[np.float64] | xr.DataArray): Azimuth angle.
        elevation (np.ndarray[np.float64] | xr.DataArray): Elevation angle.

    Raises:
        ValueError: Azimuth angle is not constant.

    Returns:
        tuple[np.ndarray[np.float64], np.ndarray[np.float64]]: x, y coordinates in cartessian plane [units as `ranges`].
    """    

    if isinstance(ranges, xr.DataArray):
        ranges = ranges.values
    if isinstance(azimuth, xr.DataArray):
        azimuth = azimuth.values
    if isinstance(elevation, xr.DataArray):
        elevation = elevation.values

    # Check elevation angle is constant    
    if not np.allclose(azimuth, azimuth[0], atol=0.1):
        raise ValueError("Azimuth angle is not constant.")
    
    x = ranges[:, np.newaxis]  * np.cos(np.deg2rad(elevation))
    y = ranges[:, np.newaxis]  * np.sin(np.deg2rad(elevation))

    return x, y

def spherical_to_cartessian(
    ranges: np.ndarray[np.float64] | xr.DataArray,    
    azimuth: np.ndarray[np.float64] | xr.DataArray,
    elevation: np.ndarray[np.float64] | xr.DataArray,
    ) -> tuple[np.ndarray[np.float64], np.ndarray[np.float64]]:
    """Convert RHI coordinates to cartessian coordinates

    Args:
        ranges (np.ndarray[np.float64] | xr.DataArray): Range of the radar.
        azimuth (np.ndarray[np.float64] | xr.DataArray): Azimuth angle.
        elevation (np.ndarray[np.float64] | xr.DataArray): Elevation angle.

    Raises:
        ValueError: Azimuth angle is not constant.

    Returns:
        tuple[np.ndarray[np.float64], np.ndarray[np.float64]]: x, y coordinates in cartessian plane [units as `ranges`].
    """    

    if isinstance(ranges, xr.DataArray):
        ranges = ranges.values
    if isinstance(azimuth, xr.DataArray):
        azimuth = azimuth.values
    if isinstance(elevation, xr.DataArray):
        elevation = elevation.values

    # # Check elevation angle is constant    
    # if not np.allclose(azimuth, azimuth[0], atol=0.1):
    #     raise ValueError("Azimuth angle is not constant.")
    
    # Convert azimth to speherical coordinates convention
    theta = 90 - elevation
    phi = convert_azimuth_y_to_x(azimuth, direction='clockwise')

    x = ranges[:, np.newaxis]  * np.sin(np.deg2rad(theta)) * np.cos(np.deg2rad(phi))
    y = ranges[:, np.newaxis]  * np.sin(np.deg2rad(theta)) * np.sin(np.deg2rad(phi))
    z = ranges[:, np.newaxis]  * np.cos(np.deg2rad(theta))

    return ranges, theta, phi, x, y, z

def convert_azimuth_y_to_x(angles_y, direction='clockwise'):
    """
    Convert an array of azimuth angles from being with respect to the y-axis to the x-axis.

    Parameters:
    angles_y (np.ndarray): Array of azimuth angles in degrees with respect to the y-axis.
    direction (str): Direction of measurement ('clockwise' or 'anticlockwise').

    Returns:
    np.ndarray: Array of azimuth angles in degrees with respect to the x-axis.
    """
    if direction == 'clockwise':
        angles_x = (90 - angles_y) % 360
    elif direction == 'anticlockwise':
        angles_x = (angles_y - 90) % 360
    else:
        raise ValueError("Direction must be 'clockwise' or 'anticlockwise'")
    
    return angles_x

def histogram_intersection(histogram1: np.ndarray[Any, np.float64], histogram2: np.ndarray[Any, np.float64], bins: np.ndarray[Any, np.float64]) -> float:
    """Compute the histogram intersection between two histograms.

    Args:

        - histogram1 (_type_): Histogram 1.
        - histogram2 (_type_): Histogram 2.
        - bins (_type_): Bins of the histograms.

    Returns:

        - float: 
    """    
    bins = np.diff(bins)
    sm = 0
    for i in range(len(bins)):
        sm += min(bins[i] * histogram1[i], bins[i] * histogram2[i])
    return sm

def mergeChirps_LV0(data):
# this function merges the chirps of the dataset. I am doing this, because the non-polarimetric X and Ka-Band don't have a chirps and therefore we only have one fixed range-resolution
# input: data as xarray-dataset, if you set spec=True, the spectral data will be merged.
    try:
        chirpNum = data.ChirpNum.values[0]
    except:
        chirpNum = data.ChirpNum.values

    try:
        if len(data.DoppLen) > 3:
            maxVel = data.MaxVel.values[0]
            doppLen = data.DoppLen.values[0]
        else:
            maxVel = data.MaxVel.values
            doppLen = data.DoppLen.values
    except:
        maxVel = data.MaxVel.values
        doppLen = data.DoppLen.values

    dv_vec = np.empty(chirpNum)
    maxRange = [max(data.C1Range), max(data.C2Range), max(data.C3Range)]
    for chirp in range(chirpNum):
        ChRange = 'C{chirp}Range'.format(chirp=chirp+1)
        ChHSpec = 'C{chirp}HSpec'.format(chirp=chirp+1)
        ChVSpec = 'C{chirp}VSpec'.format(chirp=chirp+1)
        ChReVHSpec = 'C{chirp}ReVHSpec'.format(chirp=chirp+1)
        ChHNoisePow = 'C{chirp}HNoisePow'.format(chirp=chirp+1)
        ChVNoisePow = 'C{chirp}VNoisePow'.format(chirp=chirp+1)
        ChVel = 'C{chirp}Vel'.format(chirp=chirp+1)
        
        #- calculate noise density:
        breakpoint()
        NoiseDensV = data[ChVNoisePow]/doppLen[chirp]
        NoiseDensH = data[ChHNoisePow]/doppLen[chirp]
        #- now we need to decompose the VSpec, because this is not actually the vert. Spectrum but saved as a composite of H and V
        data[ChVSpec] = 4*data[ChVSpec] - data[ChHSpec] - 2*data[ChReVHSpec]
        #- there was a software mistake, so the noise of ReHV was not stored correctly. Therefore Alexander suggested to use s SNR threshold of 10dB, otherwise the data will be masked. For this we need to calculate SNR: SNR = signal power/noise power. In order to calculate the correct values, we need to mask all values below -90 dBZ.
        specThreshold = 10**(-90/10)
        data[ChHSpec] = data[ChHSpec].where(data[ChHSpec]>specThreshold,np.NaN)
        data[ChVSpec] = data[ChVSpec].where(data[ChVSpec]>specThreshold,np.NaN)
        NoisePowH = data[ChHSpec].count(dim=ChVel)*NoiseDensH
        NoisePowV = data[ChVSpec].count(dim=ChVel)*NoiseDensV
        SignalPowH = data[ChHSpec].sum(dim=ChVel)
        SignalPowV = data[ChVSpec].sum(dim=ChVel)
        data['SNR_H'] = SignalPowH / NoisePowH
        data['SNR_V'] = SignalPowV / NoisePowV
        #-- now also make it spectral (so spectral ZDR against SNR in respective bin (we see also some high ZDR values at the left (fast) edge, which should not be there)) and then the data should be masked with sSNR > 10dB aswell
        data['sSNR_H'] = data[ChHSpec] / NoiseDensH
        data['sSNR_V'] = data[ChVSpec] / NoiseDensH 
        #-- it is the easiest to calculate Z,ZDR,ZDP at this stage, because once the chirps are merged, it is difficult to sort out the dv needed to integrate over the spectrum...
        data['ZH'] = data[ChHSpec].sum(dim=ChVel)
        data['ZV'] = data[ChVSpec].sum(dim=ChVel)
        data['ZDR'] = 10*np.log10(data['ZH'])-10*np.log10(data['ZV'])
        data['ZDP'] = data['ZH']-data['ZV']
        #- because the different chirps have a different Doppler resolution, we also need to regrid along that axis:
        velData = np.linspace(-maxVel[chirp], maxVel[chirp], doppLen[chirp], dtype=np.float32) # in the dataformat, the dopplervelocity was not assigned yet, but rather it is stored as maxVel and doppLen which you then need to manually assign to the doppler Vel coordinate
        dv_diff = np.diff(velData) # since we regrid along the doppler axis, we need to divide the regridded doppler spectra by dv
        #print(dv_diff)
        dv_vec[chirp] = dv_diff[0]
        data = data.assign({ChVel:velData})
        velRef = np.linspace(-maxVel[0], maxVel[0], doppLen[0], dtype=np.float32)# just use the Doppler velocity from the smallest chirp
        data = data.reindex({ChVel:velRef}, method = 'nearest', tolerance = 0.05) # regrid
        data[ChVSpec] = data[ChVSpec]/dv_vec[chirp]
        data[ChHSpec] = data[ChHSpec]/dv_vec[chirp]
        data[ChReVHSpec] = data[ChReVHSpec]/dv_vec[chirp]
        #- now we can rename the variables to without the chirps and then we merge the datasets along the range coordinate
        dataCh = data[[ChVSpec,ChHSpec,ChHNoisePow,ChVNoisePow,'ZH','ZV','ZDR','ZDP','SNR_H','SNR_V','sSNR_H','sSNR_V']]
        dataCh = dataCh.rename({ChRange:'range',ChVSpec:'VSpec',ChHSpec:'HSpec',ChHNoisePow:'HNoisePow',ChVNoisePow:'VNoisePow',ChVel:'Vel'})
        if chirp==0:
            finalData = dataCh
        else:
            finalData = xr.concat([finalData,dataCh],dim='range')
    dv =  xr.DataArray(dv_vec,dims=('Chirp'))
    temp = xr.Dataset({'Azm':data.Azm,
                    'Elv':data.Elv,
                    'RangeRes':data.RangeRes,
                    #'delRange':delRange,
                    'dv':dv,
                    'maxVel':maxVel,
                    'doppLen':doppLen,
                    'maxRange':maxRange
                    })
    
    finalData = xr.merge([finalData, temp])
    return finalData 


# This Python script will be executed from within the main lidar_correction_ghk.py
# Probably it will be better in the future to let the main script rather read a conguration file,
# which might improve the portability of the code within an executable.
# Due to problems I had with some two letter variables, most variables are now with at least
# three letters mixed small and capital.
# To be used with lidar_correction_ghk.py ver. 0.9.8e and larger

# Do you want to calculate the errors? If not, just the GHK-parameters are determined.
Error_Calc = False

# Header to identify the lidar system
EID = "gr"				# Earlinet station ID 
LID = "mulhacen" # Additional lidar ID (short descriptive text)
print("    Lidar system :", EID, ", ", LID)

# +++ IL Laser and +-Uncertainty
Qin, dQin, nQin = 0.995, 0.005,  1	# second Stokes vector parameter; default 1 => linear polarization  0.999 => LDR = 0.0005
Vin, dVin, nVin = 0.0, 0.0,  0	# fourth Stokes vector parameter; default 0  => corresponds to LDR 0.0005 with DOP 1
RotL, dRotL, nRotL 	= 7.1, 	0.3, 	1	#alpha; rotation of laser polarization in degrees; default 0

# +++ ME Emitter optics and +-Uncertainty;  default = no emitter optics
DiE, dDiE, nDiE 	= 0.0, 	0.1, 	0	#  Diattenuation; default 0
TiE 		        = 1.0		        # Unpolarized transmittance; default 1
RetE, dRetE, nRetE 	= 0., 	180., 	0	# Retardance in degrees; default 0
RotE, dRotE, nRotE 	= 0., 	1., 	0	# beta: Rotation of optical element in degrees; default 0

# +++ MO Receiver optics including telescope 
DiO,  dDiO, nDiO 	= 0.88, 	0.1,    1	# Diattenuation; default 0
TiO 				= 1.0   	# Unpolarized transmittance; default 1 				
RetO, dRetO, nRetO 	= 0., 	180., 	0	# Retardance in degrees; default 0  
RotO, dRotO, nRotO 	= 0., 	0.1, 	0	#gamma: Rotation of the optical element in degrees; default 0
 
# +++++ PBS MT Transmitting path defined with TS, TP, PolFilter extinction ratio ERaT, and +-Uncertainty
#   --- Polarizing beam splitter transmitting path
TP,   dTP, nTP	 	= 0.95,	0.01,   1 # transmittance of the PBS for parallel polarized light
TS,   dTS, nTS	 	= 0.005,	0.001,   1 # transmittance of the PBS for cross polarized light
RetT, dRetT, nRetT	 = 0.0,	180., 	0 # Retardance in degrees
#   --- Pol.Filter behind transmitted path of PBS
ERaT, dERaT, nERaT	 = 1e-3, 5e-4, 0 # Extinction ratio
RotaT, dRotaT, nRotaT = 0.,   1.,    0 # Rotation of the Pol.-filter in degrees; usually close to 0° because TP >> TS, but for PollyXTs it can also be close to 90°
#   --
TiT = 0.5 * (TP + TS)    # do not change this
DiT = (TP-TS)/(TP+TS)    # do not change this
DaT = (1-ERaT)/(1+ERaT)    # do not change this
TaT = 0.5*(1+ERaT)    # do not change this

# +++++ PBS MR Reflecting path defined with RS, RP, and cleaning PolFilter extinction ratio ERaR  and +-Uncertainty
#   ---- for PBS without absorption the change of RS and RP must depend on the change of TP and TS. Hence the values and uncertainties are not independent.
RS_RP_depend_on_TS_TP = True
#   --- Polarizing beam splitter reflecting path
if(RS_RP_depend_on_TS_TP): 
    RP, dRP, nRP        = 1-TP,  0.00, 0    # do not change this
    RS, dRS, nRS        = 1-TS,  0.00, 0    # do not change this
else:
    RP, dRP, nRP        = 0.05,  0.01, 1    # change this if RS_RP_depend_on_TS_TP = False; reflectance of the PBS for parallel polarized light
    RS, dRS, nRS        = 0.98,  0.01, 1    # change this if RS_RP_depend_on_TS_TP = False; reflectance of the PBS for cross polarized light
RetR, dRetR, nRetR	    = 0.0,   180., 0    # Retardance in degrees
#   --- Pol.Filter behind reflected path of PBS
ERaR, dERaR, nERaR	  = 0.001, 0.001, 1 # Extinction ratio
RotaR, dRotaR, nRotaR = 90.,   1.,    1  # Rotation of the Pol.-filter in degrees; usually 90° because RS >> RP, but for PollyXTs it can also be 0°
#   --
TiR = 0.5 * (RP + RS)    # do not change this
DiR = (RP-RS)/(RP+RS)    # do not change this
DaR = (1-ERaR)/(1+ERaR)    # do not change this
TaR = 0.5*(1+ERaR)    # do not change this
# NEW --- Additional ND filter transmission (attenuation) during the calibration (only important for statistical errors)
TCalT, dTCalT, nTCalT  = 1, 0, 0		# transmitting path, default 1, 0, 0
TCalR, dTCalR, nTCalR = 1, 0, 0		# reflecting path, default 1, 0, 0

# +++ Orientation of the PBS with respect to the reference plane (see Improvements_of_lidar_correction_ghk_ver.0.9.8_190124.pdf)
#    Y = +1: polarisation in reference plane is finally transmitted, 
#    Y = -1: polarisation in reference plane is finally reflected. 
Y = -1.

# +++ Calibrator 
#  --- Calibrator Type used; defined by matrix values below
TypeC = 3	#Type of calibrator: 1 = mechanical rotator; 2 = hwp rotator (fixed retardation); 3 = linear polarizer; 4 = qwp; 5 = circular polarizer; 6 = real HWP calibration +-22.5°
#  --- Calibrator Location
LocC = 3 #location of calibrator: 1 = behind laser; 2 = behind emitter; 3 = before receiver; 4 = before PBS
#    --- MC Calibrator parameters
if TypeC == 1:  #mechanical rotator
	DiC, dDiC, nDiC 	= 0., 	0., 	0	# Diattenuation
	TiC = 1.
	RetC, dRetC, nRetC 	= 0., 	0., 	0	# Retardance in degrees
	RotC, dRotC, nRotC 	= 0., 	0.1, 	1	#constant calibrator rotation offset epsilon
	# Rotation error without calibrator: if False, then epsilon = 0 for normal measurements	
	RotationErrorEpsilonForNormalMeasurements = True	# 	is in general True for TypeC == 1 calibrator
elif TypeC == 2:   # HWP simulated by rotator without retardance!
	DiC, dDiC, nDiC 	= 0., 	0., 	0	# Diattenuation; ideal 0.0
	TiC = 1.
	RetC, dRetC, nRetC 	= 180., 0., 	0	# Retardance in degrees
	#NOTE: use here twice the HWP-rotation-angle
	RotC, dRotC, nRotC 	= 0.0, 	0.1, 	1	#constant calibrator rotation offset epsilon
	RotationErrorEpsilonForNormalMeasurements = True	# 	is in general True for TypeC == 2 calibrator
elif TypeC == 3:   # linear polarizer calibrator. Diattenuation DiC = (1-ERC)/(1+ERC); ERC = extinction ratio of calibrator
	DiC, dDiC, nDiC 	= 0.9998, 0.00019, 1	# Diattenuation; ideal 1.0
	TiC = 0.4	# ideal 0.5
	RetC, dRetC, nRetC 	= 0., 	180., 	0	# Retardance in degrees
	RotC, dRotC, nRotC 	= 0.0, 	0.1, 	0	#constant calibrator rotation offset epsilon
	RotationErrorEpsilonForNormalMeasurements = False	# 	is in general False for TypeC == 3 calibrator
elif TypeC == 4:   # QWP calibrator
	DiC, dDiC, nDiC 	= 0.0, 	0., 	0	# Diattenuation; ideal 0.0
	TiC = 1.0	# ideal 0.5
	RetC, dRetC, nRetC 	= 90., 	0., 	0	# Retardance in degrees
	RotC, dRotC, nRotC 	= 0.0, 	0.1, 	1	#constant calibrator rotation offset epsilon
	RotationErrorEpsilonForNormalMeasurements = False	# 	is  False for TypeC == 4 calibrator
elif TypeC == 6:   # real half-wave plate rotator calibration at +-22.5°  => rotated_diattenuator_X22x5deg.odt
	DiC, dDiC, nDiC 	= 0., 	0., 	0	# Diattenuation; ideal 0.0
	TiC = 1.
	RetC, dRetC, nRetC 	= 180., 0., 	0	# Retardance in degrees
    #Note: use real HWP angles here
	RotC, dRotC, nRotC 	= 0.0, 	0.1, 	1	#constant calibrator rotation offset epsilon
	RotationErrorEpsilonForNormalMeasurements = True	# 	is in general True for TypeC == 6 calibrator
else:
    raise ValueError('calibrator not implemented yet')

# --- LDRCal assumed atmospheric linear depolarization ratio during the calibration measurements in calibration range with almost clean air (first guess)
LDRCal,dLDRCal,nLDRCal= 0.15, 0.05, 1     # spans most of the atmospheric depolarisation variability 
# LDRCal,dLDRCal,nLDRCal= 0.009, 0.005, 1     # spans the interference filter influence 
 
# ====================================================
# NOTE: there is no need to change anything below.
# ====================================================
# !!! don't change anything in this section !!!
bPlotEtax = False    # plot error histogramms for Etax
# NEW *** Only for signal noise errors *** 
nNCal = 0           # error nNCal, calibration signals: one-sigma (fixed) in nNCal steps to left and right
nNI   = 0           # error nNI, 0° signals: one-sigma (fixed) in nNI steps to left and right; NI signals are calculated from NCalT and NCalR in main programm, but noise is assumed to be independent.

#   --- number of photon counts in the signal summed up in the calibration range during the calibration measurements
NCalT = 28184		# default 1e6, assumed the same in +45° and -45° signals; counts with ND-filter TCalT 
NCalR = 28184		# default 1e6, assumed the same in +45° and -45° signals; counts with ND-filter TCalR 
NILfac = 2          # (relative duration (laser shots) of standard (0°) measurement to calibration measurements) * (range of std. meas. smoothing / calibration range); example: 100000#/5000# * 100/1000 = 2 
                    #  LDRmeas below will be used to calculate IR and IT of 0° signals.
# calculate signal counts only from parallel 0° signal assuming the same electronic amplification in both channels; overwrites above values
CalcFrom0deg = True
NI = 1e5 #number of photon counts in the parallel 0°-signal 
    
if(CalcFrom0deg):
    # either eFactT or eFacR is = 1 => rel. amplification
	eFacT = 1                     			# rel. amplification of transmitted channel, approximate values are sufficient; def. = 1
	eFacR = 10                    			# rel. amplification of reflected channel, approximate values are sufficient; def. = 1
	NILfac = 2          					# (relative duration (laser shots) of standard (0°) measurement to calibration measurements) * (range of std. meas. smoothing / calibration range); example: 100000#/5000# * 100/1000 = 2 

	NCalT = NI / NILfac * TCalT * eFacT		# photon counts in transmitted signal during calibration 
	NCalR = NI / NILfac * TCalR * eFacR	    # photon counts in reflected signal during calibration 
                        #  LDRmeas below will be used to calculate IR and IT of 0° signals.
# NEW *** End of signal noise error parameters ***  


# --- LDRtrue for simulation of measurement => LDRsim
LDRtrue = 0.004
LDRtrue2 = 0.004

# --- measured LDRm will be corrected with calculated parameters GHK
LDRmeas = 0.3

# --- this is just for correct transfer of the variables to the main file 
Qin0, dQin, nQin = Qin, dQin, nQin
Vin0, dVin, nVin = Vin, dVin, nVin
RotL0, dRotL, nRotL = RotL, dRotL, 	nRotL 
# Emitter
DiE0,  dDiE,  nDiE  = DiE,  dDiE, 	nDiE  
RetE0, dRetE, nRetE = RetE, dRetE, 	nRetE 
RotE0, dRotE, nRotE = RotE, dRotE, 	nRotE 
# Receiver
DiO0,  dDiO,  nDiO  = DiO,  dDiO, 	nDiO  
RetO0, dRetO, nRetO = RetO, dRetO, 	nRetO 
RotO0, dRotO, nRotO = RotO, dRotO, 	nRotO 
# Calibrator
DiC0,  dDiC,  nDiC  = DiC,  dDiC, 	nDiC  
RetC0, dRetC, nRetC = RetC, dRetC, 	nRetC 
RotC0, dRotC, nRotC = RotC, dRotC, 	nRotC 
# PBS
TP0,   dTP,   nTP   = TP,   dTP, 	nTP   
TS0,   dTS,   nTS   = TS,   dTS, 	nTS 
RetT0, dRetT, nRetT	= RetT, dRetT, nRetT

ERaT0, dERaT, nERaT	= ERaT, dERaT, nERaT
RotaT0,dRotaT,nRotaT= RotaT,dRotaT,nRotaT

RP0,   dRP,   nRP   = RP,   dRP,   nRP
RS0,   dRS,   nRS   = RS,   dRS,   nRS
RetR0, dRetR, nRetR	= RetR, dRetR, nRetR

ERaR0, dERaR, nERaR	= ERaR, dERaR, nERaR
RotaR0,dRotaR,nRotaR= RotaR,dRotaR,nRotaR

LDRCal0,dLDRCal,nLDRCal=LDRCal,dLDRCal,nLDRCal


from pathlib import Path
from gfatpy.utils.io import read_yaml

SCC_PLOT_INFO = read_yaml(Path(__file__).parent.absolute() / "info.yml")
PLOT_INFO = read_yaml(Path(__file__).parent.parent.parent.absolute() / "plot" / "info.yml")

__all__ = ["SCC_PLOT_INFO", "PLOT_INFO"]



import pathlib
from datetime import datetime
from typing import List, Optional, Tuple
from pathlib import Path

import typer

from gfatpy.lidar.utils.types import LidarName, MeasurementType, Telescope
from gfatpy.lidar.nc_convert.converter import convert_nc_by_date
from gfatpy.lidar.reader import reader_xarray
from gfatpy.lidar.quality_assurance.rayleigh_fit import rayleigh_fit_from_file


app = typer.Typer(no_args_is_help=True)


@app.command(
    help="Converts raw lidar data to l1 data",
    no_args_is_help=True,
)
@app.command(no_args_is_help=True)
def nc_convert(
    lidar_name: LidarName = typer.Option(..., "--lidar", "-l"),
    initial_date: datetime = typer.Option(..., "--initial-date", "-i"),
    data_dir: Path = typer.Option(
        ..., "--data-dir", "-d", readable=True, dir_okay=True, file_okay=False
    ),
    output_dir: Path = typer.Option(
        ..., "--output-dir", "-o", writable=True, dir_okay=True, file_okay=False
    ),
    measurement_type: Optional[MeasurementType] = typer.Option(
        None, "--measurement-type", "-t"
    ),
    telescope: Optional[Telescope] = typer.Option(Telescope.xf, "--telescope")
    # final_date: Optional[datetime] = typer.Option(None, "--final-date", "-f"), TODO: If necessary implement this
):
    if telescope is None:
        telescope = Telescope.xf

    convert_nc_by_date(
        lidar_name=lidar_name,
        output_dir=output_dir,
        date=initial_date,
        data_dir=data_dir,
        measurement_type=measurement_type,
        telescope=telescope,
    )


@app.command(no_args_is_help=True)
def plot(
    lidar_name: LidarName = typer.Option(..., "--lidar", "-l"),
    initial_date: datetime = typer.Option(..., "--initial-date", "-i"),
    final_date: Optional[datetime] = typer.Option(None, "--final-date", "-f"),
):
    typer.echo(f"{lidar_name}")
    typer.echo(f"{initial_date}")
    typer.echo(f"{final_date}")


@app.command(no_args_is_help=True)
def reader(
    filelist: pathlib.Path = typer.Option(..., "--file-list", "-f"),
    initial_date: datetime = typer.Option(..., "--initial-date", "-i"),
    final_date: Optional[datetime] = typer.Option(None, "--final-date", "-f"),
    initial_range: Optional[float] = typer.Option(0.0, "--initial-range", "-g"),
    final_range: Optional[float] = typer.Option(30000.0, "--final-range", "-t"),
    percentage_required: Optional[float] = typer.Option(
        80.0, "--percentage-required", "-p"
    ),
    _channels: Optional[str] = typer.Option([], "--channels", "-c"),
):
    if _channels is None:
        channels = []
    elif isinstance(_channels, str):
        channels = _channels.split(",")

    elif isinstance(_channels, list):
        channels = _channels
    else:
        raise ValueError("channels must be a list or a string")

    typer.echo(f"{filelist}")
    typer.echo(f"{initial_date}")
    typer.echo(f"{final_date}")
    typer.echo(f"{initial_range}")
    typer.echo(f"{final_range}")
    typer.echo(f"{percentage_required}")
    typer.echo(f"{channels}")

    reader_xarray(
        filelist,
        date_ini=initial_date,
        date_end=final_date,
        ini_range=initial_range,
        end_range=final_range,
        percentage_required=percentage_required,
        channels=channels,
    )


@app.command(help="QA Rayleigh fit", no_args_is_help=True)
def rayleigh_fit(
    file: Path = typer.Option(
        ..., "--file", "-f", readable=True, dir_okay=False, file_okay=True
    ),
    initial_hour: int = typer.Option(None, "--initial-hour", "-i"),
    duration: int = typer.Option(None, "--duration", "-d"),
    channels: Optional[List[str]] = typer.Option(None, "--channels", "-c"),
    reference_range: Optional[Tuple[float, float] | None] = typer.Option(
        None, "--reference-range", "-r"
    ),
    smooth_window: Optional[float | None] = typer.Option(None, "--smooth-window", "-s"),
    save_fig: bool = typer.Option(False, "--save-fig", "-g"),
    output_dir: Path = typer.Option(
        ...,
        "--output-dir",
        "-o",
        readable=True,
        writable=True,
        dir_okay=True,
        file_okay=False,
    ),
):
    if channels is not None:
        assert len(channels) > 0, "At least one channel is required"

    if reference_range is None:
        reference_range = (7000, 8000)

    if smooth_window is None:
        smooth_window = 250

    rayleigh_fit_from_file(
        file=file,
        initial_hour=initial_hour,
        duration=duration,
        channels=channels,
        reference_range=reference_range,
        smooth_window=smooth_window,
        output_dir=output_dir,
        save_fig=save_fig,
    )



import os, glob
import pandas as pd
from typing import Tuple

from loguru import logger


def header(file_wildcard: str) -> dict:
    """Read all types of aeronet files (*.levX, *.ONEILL_levX, *.tot_levX, and *.all). It accepts wildcard (please use '*' to create wildcard smartly).

    Args:

        - file_wildcard (str): File path wildcard.

    Raises:

        - RuntimeError: 'None or more than one file found. file_wildcard must fit only one file.'

    Return:

        - header (dict): dictionary with fields: `aeronet_version`, `aeronet_station`, `aeronet_data`, `PI_name` and `PI_email`.
    """
    fl = glob.glob(file_wildcard)

    if len(fl) != 1:
        logger.error(
            "None or more than one file found. file_wildcard must fit only one file."
        )
        raise RuntimeError

    # Check same type of file
    head = None
    if len(set([os.path.splitext(fni)[1] for fni in fl])) == 1:
        for f in fl:
            if os.path.isfile(f):
                # open file in read mode
                with open(f, "r") as myfile:
                    head = [next(myfile) for x in range(6)]

    if head is None:
        raise RuntimeError("Header not found.")

    version_, data_ = head[2].split(": ")
    header = {
        "aeronet_version": version_,
        "aeronet_station": head[1][:-1],
        "aeronet_data": data_[:-1],
        "PI_name": head[4].split("PI=")[-1].split(";")[0].replace("_", " "),
        "PI_email": head[4].split("PI Email=")[-1].split(" ")[0],
    }

    return header


def reader(file_wildcard: str) -> Tuple[dict, pd.DataFrame]:
    """Read all types of aeronet files (*.levX, *.ONEILL_levX, *.tot_levX, and *.all). It accepts wildcard (please use '*' to create wildcard smartly). It concatenates data by Datetime Index

    Args:

        - file_wildcard (str): File path wildcard.

    Raises:

        - ValueError: No file found.
        - ValueError: More than onr type of files found. Please, restrict the file_wildcard.

    Returns:
    
        - pd.DataFrame: AERONET database.
    """
    fl = glob.glob(file_wildcard)

    if len(fl) == 0:
        logger.critical("No file found.")
        raise ValueError

    # Check same type of file
    if len(set([os.path.splitext(fni)[1] for fni in fl])) > 1:
        logger.critical(
            "More than onr type of files found. Please, restrict the file_wildcard. Exit"
        )
        raise ValueError

    header_dict = header(file_wildcard)
    df = []
    for f in fl:
        if os.path.isfile(f):
            dfi = pd.read_csv(f, skiprows=6)
            try:
                dfi["Datetime"] = pd.to_datetime(
                    dfi["Date(dd:mm:yyyy)"] + " " + dfi["Time(hh:mm:ss)"],
                    format="%d:%m:%Y %H:%M:%S",
                )
                for key_ in [
                        "Date(dd:mm:yyyy)",
                        "Time(hh:mm:ss)",
                        "Day_of_Year",
                        "Day_of_Year(Fraction)",
                    ]:
                    if key_ in dfi.keys():
                        dfi = dfi.drop(key_, axis=1)
            except Exception:                
                dfi["Datetime"] = pd.to_datetime(
                    dfi["Date_(dd:mm:yyyy)"] + " " + dfi["Time_(hh:mm:ss)"],
                    format="%d:%m:%Y %H:%M:%S",
                )
                try:
                    for key_ in [
                            "Date_(dd:mm:yyyy)",
                            "Time_(hh:mm:ss)",
                            "Day_of_Year",
                            "Day_of_Year(fraction)",
                        ]:
                        if key_ in dfi.keys():
                            dfi = dfi.drop(key_, axis=1)
                except Exception:
                    pass
            dfi.set_index("Datetime", inplace=True)
        else:
            raise ValueError("read_csv failed to open.")

        df.append(dfi)

    df = pd.concat(df)

    # Remove non available properties
    for key_ in df.keys():
        if (df[key_].values == -999).all():
            df = df.drop(key_, axis=1)

    return header_dict, df



from datetime import datetime
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.utils import COMMASPACE, formatdate
import smtplib

from loguru import logger

from gfatpy.lidar.utils.utils import LIDAR_INFO


email_receiver = LIDAR_INFO['emails']
email_sender = LIDAR_INFO['email_sender']

def send_email(email_sender: dict, email_receiver: list[str], subject: str, email_content: str):
    """Send email to the specified email addresses

    Args:
        email_sender (dict): Dictionary with following keys: server, port, email_name, password.
        email_receiver (list[str]): List of email addresses to send the email to.
        email_content (str): Content of the email to be sent.

    Returns:
        _type_: _description_
    """
    
    # Check Input
    if isinstance(email_receiver, str):
        email_receiver = [email_receiver]

    logger.info("Send Email")
    date = datetime.utcnow()
    return_code = 1
    try:
        # Prepare Message
        message = MIMEMultipart()
        message["From"] = email_sender["email_name"]
        message["To"] = COMMASPACE.join(email_receiver)
        message["Date"] = formatdate(localtime=True)
        message["Subject"] = subject
        message.attach(MIMEText(email_content))

        # Connect to email server and send email
        # ssl_context = ssl.create_default_context()
        # conn = smtplib.SMTP_SSL(email_sender['server'], email_sender['port'], context=ssl_context)
        conn = smtplib.SMTP(email_sender["server"], email_sender["port"])
        conn.ehlo()
        conn.starttls()
        conn.ehlo()
        conn.login(email_sender["email_name"], email_sender["password"])
        conn.sendmail(email_sender["email_name"], email_receiver, message.as_string())
        logger.info("Email Sent")
    except:
        logger.error("Email not sent")
        return_code = 0
    finally:
        conn.quit()
    return return_code


import os
from multiprocessing import Pool

import numpy as np
import pandas as pd
from loguru import logger
from scipy import signal
import matplotlib.pyplot as plt

from gfatpy.utils import utils

""" LAYER DETECTION

    Description of Algorithm:
    -------------------------
    It takes RCS

"""


def find_peaks_1d(rcs, ranges, check=True, plot_results=False):
    """_summary_

    Args:
        rcs (_type_): _description_
        ranges (_type_): _description_
        method (_type_): _description_

    Returns:
        _type_: _description_
    """

    def check_candidate_peak(x, pk_ind):
        """_summary_

        Args:
            x (_type_): _description_
            pk_ind (_type_): _description_

        Returns:
            _type_: _description_
        """
        ok = False

        # Define profiles around the peak candidate
        coarse_width = 21
        coarse_box = pd.DataFrame(
            {"x": x[max(0, pk_ind - coarse_width) : min(len(x), pk_ind + coarse_width)]}
        )

        fine_width = 5
        fine_box = pd.DataFrame(
            {"x": x[max(0, pk_ind - fine_width) : min(len(x), pk_ind + fine_width)]}
        )

        # Rolling avg, std
        roll_w = 3
        coarse_box["avg"] = coarse_box["x"].rolling(roll_w, center=True).mean()
        coarse_box["std"] = coarse_box["x"].rolling(roll_w, center=True).std()

        # The algorithm
        fine_std = fine_box["x"].std()
        coarse_std_avg = coarse_box["std"].mean()
        coarse_std_std = coarse_box["std"].std()

        # Condition to accept candidate
        ok = fine_std > (coarse_std_avg + 1.1 * coarse_std_std)

        return ok, fine_std, coarse_std_avg, coarse_std_std

    # The Method
    logger.debug("Start Method Simple")
    method = "simple"
    # Use normalized RCS
    rcs_norm = utils.normalize(rcs)
    # Focus on intense peaks
    peak_candidates, out_dict = signal.find_peaks(rcs_norm, height=0.98)
    logger.debug("End Method Simple")

    # Check behavior at neighbourhood to decide
    if check:
        logger.debug("Start Check Peak Candidates")
        peak_ind = []
        # methodology based on standard deviation
        for pk in peak_candidates:
            ok = check_candidate_peak(rcs_norm, pk)
            if ok:
                peak_ind.append(pk)
        logger.debug("End Check Peak Candidates")
    else:
        peak_ind = peak_candidates

    if plot_results:
        f, ax = plt.subplots()
        ax.plot(rcs_norm, ranges, label="rcs norm")
        ax.plot(rcs_norm[peak_candidates], ranges[peak_candidates], "o", color="r")
        ax.legend()
        ax.set_title(method)

    return peak_ind


def find_peaks(
    rcs_2d, ranges, min_range=None, max_range=None, method="default", parallel=False
):
    """Simple Method based on scipy.signal.find_peaks
    Useful for Strong Layers

    Args:
        rcs_2d (_type_): _description_
        parallel (bool, optional): _description_. Defaults to True.

    Returns:
        layers (_type_): dictionary for each rcs profiles with layer detection info
    """

    logger.debug("Start Method Find Peaks")

    if method == "default":
        method = "simple"

    # Check rcs_2d is really 2D
    ndims = rcs_2d.ndim
    if ndims == 1:
        rcs_2d = rcs_2d[np.newaxis, :]

    # Clip Range
    if not min_range:
        min_range = ranges[0]
    if not max_range:
        max_range = ranges[-1]
    idx = np.logical_and(ranges >= min_range, ranges <= max_range)
    rcs_2d = rcs_2d[:, idx]
    ranges = ranges[idx]

    # Turn off Parallel if few profiles
    dim_p, dim_h = rcs_2d.shape
    if dim_p > 2000:
        parallel = True

    # Find layers: If many, takes first
    layers = {}
    if parallel:
        with Pool(os.cpu_count()) as pool:
            input = [(x, ranges) for x in rcs_2d]
            peaks_arr = np.array(pool.starmap(find_peaks_1d, input))
        for t, peaks in enumerate(peaks_arr):
            exist = False
            if len(peaks) > 0:
                exist = True
            layers[t] = {"exist": exist, "ranges": ranges[peaks]}
    else:
        for t in range(dim_p):
            peaks = find_peaks_1d(rcs_2d[t, :], ranges)
            exist = False
            if len(peaks) > 0:
                exist = True
            layers[t] = {"exist": exist, "ranges": ranges[peaks]}

    logger.debug("End Method Find Peaks")
    return layers


def detect_layer(
    rcs_2d, ranges, min_range=None, max_range=None, method="find_peaks", **kwargs
):
    """_summary_

    Args:
        rcs_2d (_type_): _description_
        ranges (_type_): _description_
        min_range (_type_, optional): _description_. Defaults to None.
        max_range (_type_, optional): _description_. Defaults to None.
        method (str, optional): _description_. Defaults to 'find_peaks'.
    """

    logger.info("Start Layer Detection")
    if method == "find_peaks":
        layers = find_peaks(rcs_2d, ranges, min_range=min_range, max_range=max_range)
    else:
        logger.critical("Method %s does not exist" % method)
        layers = None

    logger.info("End Layer Detection")
    return layers



""" Command line tool to convert Licel binary files to SCC NetCDF format.
"""
import argparse
import glob
import importlib
import logging
import os
from pdb import set_trace
import sys

import yaml

from .licel import LicelLidarMeasurement
from .__init__ import __version__

logger = logging.getLogger(__name__)


def create_custom_class(custom_netcdf_parameter_path, use_id_as_name=False, temperature=25., pressure=1020.,
                        licel_timezone='UTC'):
    """ This funtion creates a custom LicelLidarMeasurement subclass,
    based on the input provided by the users.

    Parameters
    ----------
    custom_netcdf_parameter_path : str
       The path to the custom channels parameters.
    use_id_as_name : bool
       Defines if channels names are descriptive or transient digitizer IDs.
    temperature : float
       The ground temperature in degrees C (default 25.0).
    pressure : float
       The ground pressure in hPa (default: 1020.0).
    licel_timezone : str
       String describing the timezone according to the tz database.

    Returns
    -------
    CustomLidarMeasurement:
       A custom sub-class of LicelLidarMeasurement
    """
    logger.debug('Reading parameter files: %s' % custom_netcdf_parameter_path)
    custom_netcdf_parameters = read_settings_file(custom_netcdf_parameter_path)

    class CustomLidarMeasurement(LicelLidarMeasurement):
        extra_netcdf_parameters = custom_netcdf_parameters

        def __init__(self, file_list=None):
            super(CustomLidarMeasurement, self).__init__(file_list, use_id_as_name=use_id_as_name,
                                                         licel_timezone=licel_timezone)

        def set_PT(self):
            ''' Sets the pressure and temperature at station level. This is used if molecular_calc parameter is
            set to 0 (i.e. use US Standard atmosphere).

            The results are stored in the info dictionary.
            '''

            self.info['Temperature'] = temperature
            self.info['Pressure'] = pressure

    return CustomLidarMeasurement


def read_settings_file(settings_path):
    """ Read the settings file.

    The file should contain python code."""
    if not os.path.isfile(settings_path):
        logging.error("The provided settings path does not correspond to a file.")
        sys.exit(1)

    dirname, basename = os.path.split(settings_path)
    sys.path.append(dirname)

    module_name, _ = os.path.splitext(basename)
    settings = importlib.import_module(module_name)
    return settings


def read_cloudmask_settings_file(settings_file_path):
    """ Read the configuration file.

    The file should be in YAML syntax."""

    if not os.path.isfile(settings_file_path):
        logging.error("Wrong path for cloudmask settings file (%s)" % settings_file_path)
        sys.exit(1)

    with open(settings_file_path) as yaml_file:
        try:
            settings = yaml.load(yaml_file)
            logging.debug("Read cloudmask settings file(%s)" % settings_file_path)
        except:
            logging.error("Could not parse YAML file (%s)" % settings_file_path)
            sys.exit(1)

    return settings


def get_corrected_measurement_id(args, n):
    """ Correct the provided measurement id, in case of multiple cloud-free periods. """
    if args.measurement_id is not None:
        order = int(args.measurement_id[-2:])
        new_no = order + n
        measurement_id = args.measurement_id[:-2] + str(new_no)
        measurement_no = args.measurement_number  # The same
    else:
        measurement_no = str(int(args.measurement_number) + n).zfill(2)
        measurement_id = None

    return measurement_id, measurement_no


def convert_to_scc(CustomLidarMeasurement, files, dark_pattern, measurement_id, measurement_number=None, output_dir=None):
    """ Convert files to SCC. """
    measurement = CustomLidarMeasurement(files)
    # Get a list of files containing dark measurements
    if dark_pattern != "":
        dark_files = glob.glob(dark_pattern)


        if dark_files:
            logger.debug("Using %s as dark measurements files!" % ', '.join(dark_files))
            measurement.dark_measurement = CustomLidarMeasurement(dark_files)
        else:
            logger.warning(
                'No dark measurement files found when searching for %s. Will not use any dark measurements.' % dark_pattern)
    try:
        measurement = measurement.subset_by_scc_channels()
    except ValueError as err:
        logger.error(err)
        sys.exit(1)

    # Save the netcdf
    logger.info("Saving netcdf")
    measurement.set_measurement_id(measurement_id)
    measurement.save_as_SCC_netcdf(output_dir=output_dir)
    logger.info("Created file %s" % measurement.scc_filename)


def main():
    # Define the command line argument
    parser = argparse.ArgumentParser(description="A program to convert Licel binary files to the SCC NetCDF format.")
    parser.add_argument("parameter_file", help="The path to a parameter file linking licel and SCC channels.")
    parser.add_argument("files",
                        help="Search pattern of Licel files. Use filename wildcards. (default './*.*')",
                        default="./*.*")
    parser.add_argument("-i", '--id_as_name',
                        help="Use transient digitizer ids as channel names, instead of descriptive names",
                        action="store_true")
    parser.add_argument("-m", "--measurement_id", help="The new measurement id", default=None)
    parser.add_argument("-n", "--measurement_number",
                        help="The measurement number for the date from 00 to 99. Used if no id is provided",
                        default="00")
    parser.add_argument("-t", "--temperature", type=float,
                        help="The temperature (in C) at lidar level, required if using US Standard atmosphere",
                        default="25")
    parser.add_argument("-p", "--pressure", type=float,
                        help="The pressure (in hPa) at lidar level, required if using US Standard atmosphere",
                        default="1020")
    parser.add_argument('-D', '--dark_measurements',
                        help="Search pattern of files containing dark measurements. Use filename wildcars, see 'files' parameter for example.",
                        default="", dest="dark_files"
                        )
    parser.add_argument('--licel_timezone', help="String describing the timezone according to the tz database.",
                        default="UTC", dest="licel_timezone",
                        )

    # Verbosity settings from http://stackoverflow.com/a/20663028
    parser.add_argument('-d', '--debug', help="Print dubuging information.", action="store_const",
                        dest="loglevel", const=logging.DEBUG, default=logging.INFO,
                        )
    parser.add_argument('-s', '--silent', help="Show only warning and error messages.", action="store_const",
                        dest="loglevel", const=logging.WARNING
                        )
    parser.add_argument('--version', help="Show current version.", action='store_true')

    args = parser.parse_args()

    # Get the logger with the appropriate level
    logging.basicConfig(format='%(levelname)s: %(message)s', level=args.loglevel)
    logger = logging.getLogger(__name__)

    # coloredlogs.install(fmt='%(levelname)s: %(message)s', level=args.loglevel)

    # Check for version
    if args.version:
        print("Version: %s" % __version__)
        sys.exit(0)

    # Get a list of files to convert
    files = glob.glob(args.files)

    # If not files found, exit
    if len(files) == 0:
        logger.error("No files found when searching for %s." % args.files)
        sys.exit(1)

    # If everything OK, proceed
    logger.info("Found {0} files matching {1}".format(len(files), os.path.abspath(args.files)))
    CustomLidarMeasurement = create_custom_class(args.parameter_file, args.id_as_name, args.temperature,
                                                 args.pressure, args.licel_timezone)

    convert_to_scc(CustomLidarMeasurement, files, args.dark_files, args.measurement_id, args.measurement_number)



from matplotlib.axes import Axes
import xarray as xr

from gfatpy.utils.plot import color_list
from gfatpy.lidar.scc.plot import SCC_PLOT_INFO, PLOT_INFO


def id2str(ID: int) -> str | None:
    """Convert any ID (channel or product) to string ID using information from SCC_PLOT_INFO.
    Args:
        hoi_configuration_ID (int): Lidar configuration ID (e.g., 783).
        ID (int): Channel or product ID (e.g., 2203).

    Returns:
        str | None: String ID (e.g., 1064fta). None means ID not found. Likely, the ID is not yet in the configuration. 
    """
    productIDs = SCC_PLOT_INFO.keys()
    if ID in productIDs:
        #Search in products
        string_ID = SCC_PLOT_INFO[ID]['string_ID']
    else:
        #Search in channels
        string_ID = None
        for productID_ in productIDs:
            channels_ = SCC_PLOT_INFO[productID_]['channels'].keys()
            if ID in channels_:
                string_ID = SCC_PLOT_INFO[productID_]['channels'][ID]
        if string_ID is None:
            raise Warning(f"ID {ID} not found in product {productID_} channels {channels_}.")
    return string_ID

def plot_angstrom(ae: dict[str: xr.DataArray], ax: Axes, ae_limits: tuple[float, float] | None = None, range_limits: tuple[float, float] | None = None) -> Axes:
    """Plot angstrom exponent.

    Args:
        ae (dict[str: xr.DataArray]): dict with angstrom exponents from `gfatpy.lidar.scc.retrieval.angstrom_exponent`
        ax (Axes): Axes handle.
        ae_limits (tuple[float, float] | None, optional): limits to the Angstrm exponent axis. Defaults to None means (-1,3).
        range_limits (tuple[float, float] | None, optional): Range limits. Defaults to None means (0,14).

    Raises:
        ValueError: Error plotting Angstrom exponent.
        ValueError: Error plotting Angstrom exponent.

    Returns:
        Axes: axis with the Angstrom exponent plots.
    """
    for ae_name in ae.keys():
        ae_da = ae[ae_name]
        ae_da['altitude'] = ae_da['altitude'] / 1e3
        if ae_da is None:
            print(f"Angstrom exponent {ae_name} is None.")
            continue
        if 'time' in ae_da.dims:
            colors = color_list(len(ae_da.time.values),cmap=PLOT_INFO['ae2cmaps'][ae_name]) 
            for idx, time_ in enumerate(ae_da.time.values):
                try:
                    ae_da.sel(time=time_).plot(y='altitude',ax=ax, label=ae_name, color=colors[idx])
                except:
                    raise ValueError(f"Error plotting Angstrom exponent")
        else:
            try:
                ae_da.plot(y='altitude',ax=ax, label=ae_name, color=PLOT_INFO['ae2color'][ae_name])
            except:
                raise ValueError(f"Error plotting Angstrom exponent")
             
    #plot vertical line at zero
    ax.axvline(x=1, color='k', linestyle='--', linewidth='0.5')
    if ae_limits is not None:
        ax.set_xlim(ae_limits)
    else:
        ax.set_xlim(-1, 3)
    if range_limits is not None:
        ax.set_ylim(range_limits)
    ax.legend()
    ax.grid()
    ax.set_xlabel(r'$AE_{\beta}$, [#]')
    return ax

     


import os
from pathlib import Path

from datetime import datetime

from gfatpy.lidar.utils.utils import LIDAR_INFO
from gfatpy.utils.io import find_nearest_filepath, read_yaml_from_info


def GHK_simulator(
    lidar_nick: str,
    calibrator: str,
    target_datetime: datetime,
    output_dir: Path,
    channel: str | None = None,
) -> list[Path]:
    """Run GHK simulator for a given lidar, calibrator and channel.

    Args:

        - lidar_nick (str): lidar nickname
        - calibrator (str): calibrator nickname (eg., "hwp", "rot", "pol")
        - target_datetime (datetime): Target datetime
        - output_dir (Path): Output directory
        - channel (str | None, optional): Lidar channel in format dddss (e.g., 532n). Defaults to None (all channels).

    Raises:

        - NotADirectoryError: Directory not found
        - FileNotFoundError: File not found

    Returns:
    
        - list[Path]: List of output paths
    """

    if not output_dir.exists() or not output_dir.is_dir():
        raise NotADirectoryError(f"{output_dir} not found.")

    if channel is None:
        INFO = read_yaml_from_info(lidar_nick, target_datetime)
        channels = INFO["GHK_channels"]
    else:
        channels = [channel]
    output_paths = []
    for channel_ in channels:
        # Ini file for each channel
        ini_ghk_dir = Path(__file__).parent.absolute() / "GHK" / "system_settings"
        ini_filepath = find_nearest_filepath(
            ini_ghk_dir,
            f"optic_input_{lidar_nick}_{calibrator}_{channel_}*.py",
            5,
            target_datetime,
            and_previous=True,
        )

        if not ini_filepath.exists():
            raise FileNotFoundError(f"Ini file not found: {ini_filepath}.")

        output_path_ = run_GHK_simulator(ini_filepath, output_dir)
        if isinstance(output_path_, Path):
            output_paths.append(output_path_)
    return output_paths


def run_GHK_simulator(ini_path: Path, output_dir: Path) -> Path:
    """Run GHK simulator for a given ini file.

    Args:

        - ini_path (Path): Input ini file
        - output_dir (Path): Output directory

    Raises:

        - FileNotFoundError: File not found

    Returns:

        - Path: Output path
    """
    depo_path = Path(__file__).parent.absolute()

    GHK_program = depo_path / "GHK" / "GHK_0.9.8h_Py3.7.py"

    output_dir.mkdir(parents=True, exist_ok=True)

    os.system(f"python {GHK_program} {ini_path} {output_dir.absolute()}")

    output_path = [*output_dir.rglob(f"*{ini_path.name.split('.')[0]}*.dat")]

    if isinstance(output_path, list) and len(output_path) == 1:
        output_path = output_path[0]
        return output_path
    else:
        raise FileNotFoundError(f"Output file not found in {output_dir}.")



from pathlib import Path

from . import converter, utils  
from .utils import DSD_INFO


__all__ = ["converter", "utils", "DSD_INFO"]

__doc__ = """Parsivel disdrometer data processing tools.

    Contributors:

    - Michael Haimerl
"""


import logging
from pdb import set_trace

import numpy as np

from .licel import LicelLidarMeasurement

logger = logging.getLogger(__name__)


class LicelCalibrationMeasurement(LicelLidarMeasurement):

    def __init__(self, plus45_files=None, minus45_files=None, use_id_as_name=False, get_name_by_order=False, licel_timezone='UTC'):
        """Class to handle depolarization calibration measurements according to the SCC.


        Parameters
        ----------
        plus45_files : list of str
           List of paths for the plus 45 files.
        minus45_files : list of str
           List of paths for the minus 45 files.
        use_id_as_name : bool
           Defines if channels names are descriptive or transient digitizer IDs.
        get_name_by_order : bool
           If True, channels are named by the order they appear.
        licel_timezone : str
           String describing the timezone according to the tz database.
        """
        # Setup the empty class
        super(LicelCalibrationMeasurement, self).__init__(use_id_as_name=use_id_as_name, licel_timezone=licel_timezone,
                                                          get_name_by_order=get_name_by_order)

        self.plus45_files = plus45_files
        self.minus45_files = minus45_files

        if plus45_files and minus45_files:
            self.files = plus45_files + minus45_files

            self.check_equal_length()
            self.read_channel_data()
            self.update()

    def subset_by_scc_channels(self):
        m = super(LicelCalibrationMeasurement, self).subset_by_scc_channels()
        m.plus45_measurement = self.plus45_measurement.subset_by_scc_channels()
        m.minus45_measurement = self.minus45_measurement.subset_by_scc_channels()
        return m

    def update(self):
        """
        Correct timescales after each update.
        """
        super(LicelCalibrationMeasurement, self).update()
        self.correct_timescales()

    def check_equal_length(self):
        """
        Check if input time series have equal lengths.
        """
        len_plus = len(self.plus45_files)
        len_minus = len(self.minus45_files)
        if len_plus != len_minus:
            raise self.UnequalMeasurementLengthError(
                            "Input timeseries have different length: %s vs %s." % (len_plus, len_minus))

    def read_channel_data(self):
        # Read plus and minus 45 measurements
        self.plus45_measurement = LicelLidarMeasurement(self.plus45_files, use_id_as_name=self.use_id_as_name,
                                                        get_name_by_order=self.get_name_by_order,
                                                        licel_timezone=self.licel_timezone)
        self.plus45_measurement.extra_netcdf_parameters = self.extra_netcdf_parameters
        self.plus45_measurement.rename_channels(suffix='_p45')

        self.minus45_measurement = LicelLidarMeasurement(self.minus45_files, use_id_as_name=self.use_id_as_name,
                                                         get_name_by_order=self.get_name_by_order,
                                                         licel_timezone=self.licel_timezone)
        self.minus45_measurement.extra_netcdf_parameters = self.extra_netcdf_parameters
        self.minus45_measurement.rename_channels(suffix='_m45')

        # Combine them in this object
        self.channels = {}
        self.channels.update(self.plus45_measurement.channels)
        self.channels.update(self.minus45_measurement.channels)

        self.raw_info = self.plus45_measurement.raw_info.copy()
        self.raw_info.update(self.minus45_measurement.raw_info)

    def correct_timescales(self):
        self.check_timescales_are_two()
        self.combine_scales()

    def check_timescales_are_two(self):
        no_timescales = len(self.variables['Raw_Data_Start_Time'])
        if no_timescales != 2:
            raise self.WrongNumberOfTimescalesError("Wrong number of timescales: %s instead of 2." % no_timescales)

    def combine_scales(self):
        start_times, end_times = self.get_ordered_timescales()
        new_start_time = start_times[0]
        new_stop_time = end_times[1]
        self.variables['Raw_Data_Start_Time'] = [new_start_time, ]
        self.variables['Raw_Data_Stop_Time'] = [new_stop_time, ]
        self.reset_timescale_id()

        self.dimensions['nb_of_time_scales'] = 1

    # def _get_custom_global_attributes(self):
    #     """
    #     NetCDF global attributes that should be included in the final NetCDF file.
    #
    #     Using the values of just p45 measurements.
    #     """
    #     return self.plus45_measurement._get_custom_global_attributes()

    def reset_timescale_id(self):
        """
        Set all timescales to 0
        :return:
        """
        timescale_dict = self.variables['id_timescale']
        self.variables['id_timescale'] = dict.fromkeys(timescale_dict, 0)

    def get_ordered_timescales(self):
        scale_start_1, scale_start_2 = self.variables['Raw_Data_Start_Time']
        scale_end_1, scale_end_2 = self.variables['Raw_Data_Stop_Time']

        if scale_start_1[0] > scale_start_2[0]:
            scale_start_1, scale_start_2 = scale_start_2, scale_start_1

        if scale_end_1[0] > scale_end_2[0]:
            scale_end_1, scale_end_2 = scale_end_2, scale_end_1

        return (scale_start_1, scale_start_2), (scale_end_1, scale_end_2)

    def add_fake_measurements(self, no_profiles, variation=0.1):
        """
        Add a number of fake measurements. This is done to allow testing with single analog profiles.

        Adds a predefined variation in each new profile.
        """
        duration = self.info['duration']
        for channel_name, channel in self.channels.items():
            base_time = list(channel.data.keys())[0]
            base_data = channel.data[base_time]

            for n in range(no_profiles):
                random_variation = base_data * (np.random.rand(len(base_data)) * 2 - 1) * variation

                new_time = base_time + n * duration
                new_data = channel.data[base_time].copy() + random_variation
                if 'ph' in channel_name:
                    new_data = new_data.astype('int')
                channel.data[new_time] = new_data

        self.update()

    def subset_photoncounting(self):
        """
        Subset photoncounting channels.
        """
        ph_channels = [channel for channel in self.channels.keys() if 'ph' in channel]
        new_measurement = self.subset_by_channels(ph_channels)
        return new_measurement

    def _get_scc_channel_variables(self):
        """
        Get a list of variables to put in the SCC.

        It can be overridden e.g. in the depolarization product class.

        Returns
        -------

        channel_variables: dict
           A dictionary with channel variable specifications.
        """
        channel_variables = \
            {'Background_Low': (('channels',), 'd'),
             'Background_High': (('channels',), 'd'),
             'LR_Input': (('channels',), 'i'),
             'DAQ_Range': (('channels',), 'd'),
             'Pol_Calib_Range_Min': (('channels',), 'd'),
             'Pol_Calib_Range_Max': (('channels',), 'd'),
             }
        return channel_variables

    class UnequalMeasurementLengthError(RuntimeError):
        """ Raised when the plus and minus files have different length.
        """
        pass

    class WrongNumberOfTimescalesError(RuntimeError):
        """ Raised when timescales are not two.
        """
        pass


class DarkLicelCalibrationMeasurement(LicelCalibrationMeasurement):

    def __init__(self, dark_files=None, use_id_as_name=False, licel_timezone='UTC'):
        """Class to handle dark files for depolarization calibration measurements according to the SCC.

        It assumes that a single sent of dark measurements will be use for both plus and minus 45 channels.

        Parameters
        ----------
        dark_files : list of str
           List of paths for the dark measurement files.
        use_id_as_name : bool
           Defines if channels names are descriptive or transient digitizer IDs.
        licel_timezone : str
           String describing the timezone according to the tz database.
        """
        # Setup the empty class
        super(DarkLicelCalibrationMeasurement, self).__init__(dark_files, dark_files,
                                                              use_id_as_name=use_id_as_name,
                                                              licel_timezone=licel_timezone)

    def correct_timescales(self):
        """ For dark measuremetns, no need to correct timescales. """
        pass



# INSERT HERE THE SYSTEM PARAMETERS
general_parameters = {
    "System": "ALHAMBRA",
    "Laser_Pointing_Angle": 3,
    "Molecular_Calc": 0,  # Use US standard atmosphere
    "Latitude_degrees_north": 37.16,
    "Longitude_degrees_east": -3.60,
    "Altitude_meter_asl": 660.0,
    "Call sign": "gr",
}

# LINK YOUR LICEL CHANNELS TO SCC PARAMETERS. USE BT0, BC0 ETC AS NAMES (AS IN LICEL FILES).
channel_parameters = {
    "BT10": {
        "channel_ID": 2123,  # 1064fta
        "Background_Low": 40000.0,
        "Background_High": 50000.0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "DAQ_Range": 100.0,
    },
    "BT11": {
        "channel_ID": 2124,  # 532npa
        "Background_Low": 40000.0,
        "Background_High": 50000.0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "DAQ_Range": 100.0,
    },
    "BC11": {
        "channel_ID": 2125,  # 532npp
        "Background_Low": 40000.0,
        "Background_High": 50000.0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "DAQ_Range": 100.0,
    },
    "BT12": {
        "channel_ID": 2126,  # 532nsa
        "Background_Low": 40000.0,
        "Background_High": 50000.0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "DAQ_Range": 100.0,
    },
    "BC12": {
        "channel_ID": 2127,  # 532nsp
        "Background_Low": 40000.0,
        "Background_High": 50000.0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "DAQ_Range": 100.0,
    },
    "BT13": {
        "channel_ID": 2128,  # 355npa
        "Background_Low": 40000.0,
        "Background_High": 50000.0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "DAQ_Range": 100.0,
    },
    "BC13": {
        "channel_ID": 2129,  # 355npp
        "Background_Low": 40000.0,
        "Background_High": 50000.0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "DAQ_Range": 100.0,
    },
    "BT14": {
        "channel_ID": 2130,  # 355nsa
        "Background_Low": 40000.0,
        "Background_High": 50000.0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "DAQ_Range": 100.0,
    },
    "BC14": {
        "channel_ID": 2131,  # 355nsp
        "Background_Low": 40000.0,
        "Background_High": 50000.0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "DAQ_Range": 100.0,
    },
}



#!/usr/bin/env python
"""
SCC GFAT

Functionality: Chain Process of Lidar Measurements from Raw format to SCC server processing
    1. Convert Lidar Raw Data (0a) into SCC format
    2. Make some plots of the SCC format data
    3. Upload SCC format data to SCC server where data is processed
    4. Download Data processed in SCC server
    5. Make some plots of the processed data downloaded from SCC server

Usage:
run scc.py
    -i YYYYMMDD
    -d DATA_PATH [DATA_PATH/LIDAR/...]
    -c CAMPAIGN CONFIG FILE JSON

"""

import os
from pathlib import Path
import sys
import glob
import importlib
import shutil
import re
import argparse
from distutils.dir_util import mkpath
import itertools
import zipfile
import numpy as np
import xarray as xr
import pandas as pd
import datetime as dt
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.ticker import MultipleLocator, AutoMinorLocator
import json
import logging
from multiprocessing import Pool
import pickle
import smtplib, ssl
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.utils import COMMASPACE, formatdate
from pdb import set_trace
from gfatpy.lidar.scc.io import move2odir
from gfatpy.lidar.scc.plot import plot_scc_input, plot_scc_output
from gfatpy.lidar.scc.utils import (
    SCC_INFO,    
    TimeoutException,
    apply_pc_peak_correction,
    check_scc_output_inlocal,
    get_campaign_config,
    get_info_from_measurement_files,
    get_scc_code_from_measurement_folder,
    get_campaign_info,
    get_scc_config,
    getTP,
)

""" 3rd party code dependencies """
from gfatpy.lidar.utils.utils import LIDAR_INFO

MODULE_DIR = os.path.dirname(sys.modules[__name__].__file__)
sys.path.insert(0, MODULE_DIR)
from atmospheric_lidar.scripts import licel2scc  # copied to utils_gfat
import scc_access as sa  # copied to utils_gfat

# import gfat_config


""" """
__author__ = "Bravo-Aranda, Juan Antonio"
__license__ = "GPL"
__version__ = "1.0"
__maintainer__ = "Bravo-Aranda, Juan Antonio"
__email__ = "jabravo@ugr.es"
__status__ = "Production"


""" directories of interest  """
# Root Directory (in NASGFAT)  according to operative system
# DATA_DN = gfat_config.DATA_DN

# Source Code Dir
MODULE_DN = Path(__file__).parent

# SCC Config Directory
SCC_CONFIG_DN = MODULE_DN / "scc_configFiles"

# Run Dir
RUN_DN = os.getcwd()


""" Common Variables along the module """
STATION_ID = "gra"

LIDAR_SYSTEMS = {
    "ALHAMBRA": {"nick": "alh", "code": ""},
    "MULHACEN": {"nick": "mhc", "code": ""},
}
MEASUREMENT_TYPES = ["RS", "HF"]


""" SCC Server Connection Settings """
SCC_SERVER_SETTINGS = SCC_INFO["server_settings"]

""" logging  """
log_formatter = logging.Formatter(
    "%(levelname)s: %(funcName)s(). L%(lineno)s: %(message)s"
)
debug = True
logger = logging.getLogger(__name__)
if logger.hasHandlers():
    logger.handlers.clear()
if debug:
    handler = logging.StreamHandler(sys.stdout)
else:
    log_fn = os.path.join(
        RUN_DN, "scc_%s.log" % dt.datetime.utcnow().strftime("%Y%m%dT%H%M")
    )
    handler = logging.FileHandler(log_fn)
handler.setFormatter(log_formatter)
logger.setLevel(logging.INFO)
logger.addHandler(handler)
logger.propagate = False


""" Email 
    From a gmail account. Permission must be granted previously.
    At manage account: less secure apps, enable
"""
# TODO: cambiar a INFO.yml y crear reader
email_sender = {
    "server": "smtp.gmail.com",  # "smtp.ugr.es",
    "port": 587,
    "sender_email": "controlador.mulhacen@gmail.com",
    "password": "lidarceama12345",
}
# email_sender = {
#                "server": "smtp.ugr.es", #"smtp.gmail.com", #"smtp.ugr.es",
#                "port": 587,
#                "sender_email": "icenet@ugr.es", #"controlador.mulhacen@gmail.com",
#                "password": "ib3r1c0", # "lidarceama12345"
#                }
email_receiver = ["dbp@ugr.es", "mjgranados@ugr.es", "rascado@ugr.es", "jabravo@ugr.es"]
# email_receiver = ["dbp@ugr.es"]


def send_email(email_sender, email_receiver, email_content):
    """ """
    # Check Input
    if isinstance(email_receiver, str):
        email_receiver = [email_receiver]

    logger.info("Send Email")
    date = dt.datetime.utcnow()
    return_code = 1
    try:
        # Prepare Message
        message = MIMEMultipart()
        message["From"] = email_sender["sender_email"]
        message["To"] = COMMASPACE.join(email_receiver)
        message["Date"] = formatdate(localtime=True)
        message["Subject"] = "SCC ERROR [%s UTC]]" % date.strftime("%Y%m%d_%H%M")
        message.attach(MIMEText(email_content))

        # Connect to email server and send email
        # ssl_context = ssl.create_default_context()
        # conn = smtplib.SMTP_SSL(email_sender['server'], email_sender['port'], context=ssl_context)
        conn = smtplib.SMTP(email_sender["server"], email_sender["port"])
        conn.ehlo()
        conn.starttls()
        conn.ehlo()
        conn.login(email_sender["sender_email"], email_sender["password"])
        conn.sendmail(email_sender["sender_email"], email_receiver, message.as_string())
        logger.info("Email Sent")
    except:
        logger.error("Email not sent")
        return_code = 0
    finally:
        conn.quit()
    return return_code


"""
Processor Functions
"""


def process_day(
    lidar_name: str,
    date_str: str,
    meas_type: str,
    scc_campaign_config: int,
    process: int,
    mode: int,
    raw_dir: Path,
    scc_dir: Path,
    check_in_server: bool = False,
):
    """[summary]

    Parameters
    ----------
    date_str: str, YYYYMMDD

    Returns
    -------
    status: bool
        True: all ok; False: something went totally or partially (slot) wrong
    msg: str
        output message. Especially useful if errors

    """
    logger.info("Start Process day: %s" % date_str)

    msg = ""

    # TODO: check input variables

    # date info
    date_dt = dt.datetime.strptime(date_str, "%Y%m%d")
    i_year, i_month, i_day = [date_dt.strftime(x) for x in ["%Y", "%m", "%d"]]
    r_year = date_dt.strftime("%y")
    if date_dt.month < 10:
        r_month = "%i" % date_dt.month
    elif date_dt.month == 10:
        r_month = "A"
    elif date_dt.month == 11:
        r_month = "B"
    elif date_dt.month == 12:
        r_month = "C"
    else:
        msg = "Error Naming Month For RS filename"
        logger.warning(msg)
        return False, msg
    r_day = date_dt.strftime("%d")
    r_date_str = f"{r_year}{r_month}{r_day}"

    """ Measurement Directories """
    day_0a_dir = raw_dir / i_year / i_month / i_day

    # Measurements Directories: RS
    rs_dns = day_0a_dir.glob(f"{meas_type}_{date_str}_*")

    # Fin RS files in previous day
    date_ytd_dt = date_dt - dt.timedelta(days=1)
    date_ytd_str = date_ytd_dt.strftime("%Y%m%d")
    y_year, y_month, y_day = [date_ytd_dt.strftime(x) for x in ["%Y", "%m", "%d"]]
    previous_dir = raw_dir / y_year / y_month / y_day
    add_rs_dns = previous_dir.glob(f"{meas_type}_{date_ytd_str}_*")

    rs_dns = itertools.chain(rs_dns, add_rs_dns)

    """ Run process for every configuration in the scc campaign config """
    campaign_name = scc_campaign_config["name"]
    lidar_configs = scc_campaign_config["lidar_config"]
    # Loop over lidar configurations within the campaign
    for lidar_config in lidar_configs:
        # Get Lidar Configuration Info
        ldr_cfg = lidar_configs[lidar_config]
        scc_id = ldr_cfg["scc"]
        hour_ini = ldr_cfg["hour_ini"]
        hour_end = ldr_cfg["hour_end"]
        hour_res = ldr_cfg["hour_res"]
        timestamp = ldr_cfg["timestamp"]
        slot_name_type = ldr_cfg["slot_name_type"]

        # SCC Config File
        if scc_id is None:
            # TODO: integrar deduccion de SCC a partir de la medida
            msg = "SCC cannot be None"
            logger.error(msg)
            return False, msg
        else:
            lidar_nick = LIDAR_INFO["metadata"]["name2nick"][lidar_name]
            scc_config_fn = SCC_CONFIG_DN / f"{lidar_nick}_parameters_scc_{scc_id}.py"

            if not scc_config_fn.exists():
                raise FileNotFoundError(f"SCC config file {scc_config_fn} not found")

        scc_config_dict = get_scc_config(scc_config_fn)

        logger.info("SCC = %d" % scc_id)

        # Measurement Files within the day
        regex_ = f"R*{r_date_str}*.*"
        rs_files = []
        for rs_dn in rs_dns:
            rs_files.extend(rs_dn.rglob(regex_))
        rs_files.sort()

        if len(rs_files) > 0:
            # Time Period of Measurements within the day (Bottleneck)
            (
                times_ini_rs,
                times_end_rs,
                time_start_rs,
                time_end_rs,
                meas_channels,
            ) = get_info_from_measurement_files(rs_files, scc_config_fn)
            # This is only for testing purposes
            # with open(os.path.join(MODULE_DN, "times.pickle"), "wb") as f:
            #   pickle.dump([times_ini_rs, times_end_rs, time_start_rs, time_end_rs, meas_channels], f)
            # with open(os.path.join(MODULE_DN, "times.pickle"), "rb") as f:
            #    times_ini_rs, times_end_rs, time_start_rs, time_end_rs, meas_channels = pickle.load(f)
            # More than 10 minutes of measurements
            if (time_end_rs - time_start_rs).total_seconds() / 60 < 10:
                msg = "Less than 10 minutes of measurements. Exit Program"
                logger.error(msg)
                return False, msg

            # Build Slots:
            logger.info("Building Slots...")
            if hour_ini is None:
                hour_ini = float(
                    time_start_rs.hour
                    + time_start_rs.minute / 60.0
                    + time_start_rs.second / 3600.0
                )
            else:
                hour_ini = float(hour_ini)
            if hour_end is None:
                if time_end_rs.day == time_start_rs.day:
                    hour_end = float(
                        time_end_rs.hour
                        + time_end_rs.minute / 60.0
                        + time_end_rs.second / 3600.0
                    )
                else:
                    hour_end = 23.99
            else:
                hour_end = float(hour_end)
            if hour_res is None:
                hour_res = float(1)
            else:
                hour_res = float(hour_res)
            if timestamp is None:
                timestamp = 1
            if slot_name_type is None:
                slot_name_type = 0
            if hour_end == hour_ini:
                hour_end = hour_ini + hour_res
            slots_ini = [
                date_dt + dt.timedelta(hours=x)
                for x in np.arange(hour_ini, hour_end, hour_res)
            ]
            slots_end = [s + dt.timedelta(hours=hour_res) for s in slots_ini]
            slots_stp = [
                s + dt.timedelta(hours=(timestamp / 2) * hour_res) for s in slots_ini
            ]
            logger.info("Done")

            # SCC input directory for the day:
            if isinstance(hour_res, int) or hour_res.is_integer():
                input_pttn = "input_%02dhoras" % hour_res
            else:
                input_pttn = "input_%dminutes" % int(hour_res * 60)
            scc_input_dn = (
                scc_dir / f"scc{scc_id}" / input_pttn / i_year / i_month / i_day
            )
            if not scc_input_dn.exists():
                scc_input_dn.mkdir(parents=True)
                logger.info("Local Directory %s CREATED" % scc_input_dn)

            # SCC Output Directory for the day
            scc_output_dn = Path(scc_input_dn.as_posix().replace("input", "output"))

            # SCC Object for Interaction withh SCC Server
            logger.info("Creating SCC object...")
            SCC_SERVER_SETTINGS[
                "output_dir"
            ] = scc_output_dn  # se tiene que llamar asi por scc_access
            scc_obj = sa.SCC(
                tuple(SCC_SERVER_SETTINGS["basic_credentials"]),
                SCC_SERVER_SETTINGS["output_dir"],
                SCC_SERVER_SETTINGS["base_url"],
            )
            logger.info("Done")

            # Loop over Slots
            # TODO: paralelizar?
            for s in range(len(slots_stp)):
                """Name Of Slot"""
                slot_number = s
                # Slot ID. For the Name of the Netcdf
                if slot_name_type == 0:  # EARLINET. operational slot_id
                    slot_id = f"{date_str}{STATION_ID}{slot_number}"
                elif slot_name_type == 1:  # SCC. timestamp
                    slot_id = f"{date_str}{STATION_ID}{slots_stp[s].hour:02d}{slots_stp[s].minute:02d}"
                elif slot_name_type == 2:  # ALTERNATIVE. timestamp + station + scc
                    slot_id = f"{date_str}{STATION_ID}{slots_stp[s].hour:02d}{slots_stp[s].minute:02d}_{scc_id}"
                else:
                    logger.warning("slot name type not defined. Set to operational")
                    slot_id = f"{date_str}{STATION_ID}{slot_number}"

                """ In Operational Mode, Check if Slot is completed so it can be processed """
                do_process_slot = True
                if mode == 0:
                    now_dt = dt.datetime.utcnow()
                    if now_dt < slots_end[s]:
                        do_process_slot = False

                """ Process Slot if Possible """
                if do_process_slot:
                    """Check Slot Exists in Local and in SCC server"""
                    # Local File for Slot
                    slot_fn = "%s.nc" % slot_id
                    scc_slot_fn = os.path.join(scc_input_dn, slot_fn)
                    if os.path.isfile(scc_slot_fn):
                        slot_in_local = True
                        logger.info("Slot %s Already exists in Local" % slot_id)
                    else:
                        slot_in_local = False
                        logger.info("Slot %s does not exist in Local" % slot_id)
                        # Slot in Server
                    if check_in_server:
                        scc_obj.login(SCC_SERVER_SETTINGS["website_credentials"])
                        meas_obj, _ = scc_obj.get_measurement(slot_id)
                        scc_obj.logout()
                        if meas_obj is not None:
                            slot_in_scc = True
                            logger.info("Slot %s Already exists in SCC" % slot_id)
                        else:
                            slot_in_scc = False
                            logger.info("Slot %s Does Not Exist in SCC" % slot_id)

                    # Local Output SCC Slot
                    scc_output_slot_dn = os.path.join(scc_output_dn, slot_id)

                    """ 0a to SCC """       
                    if np.logical_or.reduce((process == 0, process == 10)):
                        # not slot_in_local,
                        # Subset RS Files for Slot
                        ids = []
                        for (i, t_i), (j, t_e) in zip(
                            enumerate(times_ini_rs), enumerate(times_end_rs)
                        ):                            
                            slot_condition = np.logical_and(np.array(times_ini_rs) > slots_ini[s], np.array(times_ini_rs) < slots_end[s])
                            if slot_condition.any() > 0:  # IF there are files within the slot
                                logger.info("Creating SCC format for slot %s" % slot_id)
                                slot_times = times_ini_rs[slot_condition]
                                # RS Files and DC Pattern
                                rs_files_slot = [rs_files[i] for i in ids]
                                dc_files_patt = os.path.join(
                                    os.path.dirname(rs_files_slot[0]).replace(
                                        meas_type, "DC"
                                    ),
                                    "R*",
                                )
                                # if there is no operational DC, it is searched
                                if len(glob.glob(dc_files_patt)) == 0:
                                    try:
                                        logger.warning("No DC files for day %s" % date_str)
                                        logger.warning(
                                            "Search DC files within the previous 30 days"
                                        )
                                        for b_day in range(1, 30):
                                            b_date_dt = date_dt + dt.timedelta(days=-b_day)
                                            b_year, b_month, b_day = [
                                                b_date_dt.strftime(x)
                                                for x in ["%Y", "%m", "%d"]
                                            ]
                                            b_day_0a_dir = os.path.join(
                                                raw_dir, b_year, b_month, b_day
                                            )
                                            dc_files_patt = os.path.join(
                                                b_day_0a_dir, "DC*", "R*"
                                            )
                                            if len(glob.glob(dc_files_patt)) > 0:
                                                break
                                        if len(glob.glob(dc_files_patt)) > 0:
                                            dc_dns = np.unique(
                                                [
                                                    os.path.basename(os.path.dirname(x))
                                                    for x in glob.glob(dc_files_patt)
                                                ]
                                            ).tolist()
                                            dc_hours = np.asarray(
                                                [
                                                    dt.datetime.strptime(
                                                        i, "DC_%Y%m%d_%H%M"
                                                    ).hour
                                                    for i in dc_dns
                                                ]
                                            )
                                            # Daytime or Nighttime slot
                                            limit_hour = 17
                                            if (
                                                time_start_rs.hour <= limit_hour
                                            ):  # Daytimeslot
                                                ix = np.argwhere(dc_hours <= limit_hour)[0]
                                            else:  # Nighttime Slot
                                                ix = np.argwhere(dc_hours > limit_hour)[0]
                                            if len(ix) > 0:
                                                dc_dn = dc_dns[ix[0]]
                                                dc_files_patt = os.path.join(
                                                    b_day_0a_dir, dc_dn, "R*"
                                                )
                                    except Exception as e:
                                        logger.error(
                                            "No DC files found for slot %s" % slot_id
                                        )                    
                                # Take temperature and pressure from the first file:
                                temperature, pressure = getTP(rs_files_slot[0])
                                # Prepare Licel2SCC input:
                                CustomLidarMeasurement = licel2scc.create_custom_class(
                                    scc_config_fn,
                                    use_id_as_name=True,
                                    temperature=temperature,
                                    pressure=pressure,
                                )
                                set_trace()
                                # Convert from Raw to SCC Format:
                                try:
                                    licel2scc.convert_to_scc(
                                        CustomLidarMeasurement,
                                        rs_files_slot,
                                        dc_files_patt,
                                        slot_id,
                                        slot_number,
                                    )
                                except Exception as e:
                                    logger.error(
                                        "%s. Conversion from 0a to SCC not possible for slot %s"
                                        % (str(e), slot_id)
                                    )
                                    msg += "Slot %s not created. " % slot_id

                                """ Name of Netcdf File """
                                # With reverse engineering, we know that the name of the
                                # netcdf file is:
                                scc_slot_ini_fn = os.path.join(RUN_DN, slot_fn)

                                """ PC peak correction if MULHACEN """
                                if lidar_name == "MULHACEN":
                                    if os.path.isfile(scc_slot_ini_fn):
                                        logger.info("PC peak correction")
                                        pc_channels = []
                                        for c in scc_config_dict["channels"]:
                                            if c in meas_channels.keys():
                                                if (
                                                    meas_channels[c].analog_photon_string
                                                    == "ph"
                                                ):
                                                    pc_channels.append(
                                                        scc_config_dict[
                                                            "channel_parameters"
                                                        ][c]["channel_ID"]
                                                    )
                                        if len(pc_channels) > 0:
                                            scc_slot_ini_fn = apply_pc_peak_correction(
                                                [scc_slot_ini_fn], pc_channels
                                            )  # se traga una lista de archivos
                                            # Devuelve una lista de archivos. TODO: Arreglar esto
                                            scc_slot_ini_fn = scc_slot_ini_fn[0]

                                """ Move File to NAS """
                                if os.path.isfile(scc_slot_ini_fn):
                                    scc_saved = move2odir(scc_slot_ini_fn, scc_input_dn)
                                    if scc_saved:
                                        logger.info(
                                            "Created SCC Format for slot %s" % slot_id
                                        )
                                    else:
                                        logger.warning(
                                            "Error: SCC Format for slot %s not created"
                                            % slot_id
                                        )
                                else:
                                    logger.error("Slot %s not created" % slot_id)

                            else:
                                logger.warning(
                                    "No RS Files within the slot: [%s, %s]"
                                    % (
                                        slots_ini[s].strftime("%H%M"),
                                        slots_end[s].strftime("%H%M"),
                                    )
                                )

                    """ PLOT SCC INPUT """
                    if np.logical_or.reduce((process == 0, process == 20)):
                        # not slot_in_local,
                        logger.info("Plot Input SCC")
                        plot_scc_input(
                            [scc_slot_fn]
                        )  # Se traga una lista de archivos en vez de uno. TODO: arreglar esto
                    else:
                        logger.info("Plot Input SCC already done")

                    """ SCC UPLOAD AND PROCESS """
                    if np.logical_or.reduce(
                        (process == 0, process == 1, process == 30)
                    ):
                        # not slot_in_scc,
                        """UPLOAD SLOT TO SCC SERVER"""
                        if (
                            slot_in_scc
                        ):  # Delete Slot in SCC if exists: DOES NOT WORK YET
                            try:
                                logger.info(
                                    "Start Delete Slot %s from SCC server" % slot_id
                                )
                                scc_obj.login(
                                    SCC_SERVER_SETTINGS["website_credentials"]
                                )
                                deleted = scc_obj.delete_measurement(slot_id)
                                scc_obj.logout()
                                if deleted:
                                    logger.info(
                                        "Slot %s deleted from SCC server" % slot_id
                                    )
                                else:
                                    logger.error(
                                        "Slot %s NOT deleted from SCC server" % slot_id
                                    )
                                    logger.error(
                                        "Slot %s must be deleted manually from SCC server"
                                        % slot_id
                                    )
                            except Exception as e:
                                logger.error(str(e))
                                logger.error("Slot %s not deleted from SCC server")
                        try:
                            # Wrap for skip in case of upload takes longer than MAX_EXEC_TIME
                            logger.info("Start Upload Slot %s" % slot_id)
                            # old_handler = signal.signal(signal.SIGALRM, sigalrm_handler)
                            # signal.alarm(MAX_EXEC_TIME)
                            if os.path.isfile(scc_slot_fn):
                                scc_obj.login(
                                    SCC_SERVER_SETTINGS["website_credentials"]
                                )
                                scc_obj.upload_file(scc_slot_fn, scc_id)
                                meas_obj, _ = scc_obj.get_measurement(slot_id)
                                scc_obj.logout()
                                if meas_obj is not None:
                                    logger.info("End Upload Slot %s" % slot_id)
                                else:
                                    logger.info("Slot %s not uploaded" % slot_id)
                            else:
                                logger.error(
                                    "Slot %s cannot be uploaded because it does not exist"
                                    % slot_id
                                )
                        except TimeoutException:
                            logger.error(
                                "Upload file took longer than %d s" % SCC_INFO["MAX_EXEC_TIME"]
                            )  # TODO: include MAX_EXEC_TIME in info.yml
                        except Exception as e:
                            logger.error(str(e))
                            logger.error("Slot %s not uploaded" % slot_id)
                        # finally:
                        # Turn off timer, Restore handler to previous value
                        # signal.alarm(0)
                        # signal.signal(signal.SIGALRM, old_handler)

                        """ Check, afterall, if Slot is at SCC """
                        scc_obj.login(SCC_SERVER_SETTINGS["website_credentials"])
                        meas_obj, _ = scc_obj.get_measurement(slot_id)
                        scc_obj.logout()
                        if meas_obj is None:
                            logger.error(
                                "Slot %s has not been uploaded to SCC server" % slot_id
                            )
                    else:
                        logger.info("Slot %s already in SCC" % slot_id)

                    scc_output_inlocal = check_scc_output_inlocal(scc_output_slot_dn)

                    """ PROCESS AND DOWNLOAD SCC-PROCESSED """
                    if np.logical_or.reduce(
                        (
                            process == 0,
                            process == 1,
                            process == 2,
                            process == 40,
                            process == 50,
                        )
                    ):
                        # not scc_output_inlocal,
                        # not slot_in_local, not slot_in_scc,
                        try:
                            # Wrap for skip in case of upload takes longer than MAX_EXEC_TIME
                            # old_handler = signal.signal(signal.SIGALRM, sigalrm_handler)
                            # signal.alarm(MAX_EXEC_TIME)

                            # connect to scc
                            scc_obj.login(SCC_SERVER_SETTINGS["website_credentials"])

                            # get measurement object
                            meas_obj, status = scc_obj.get_measurement(slot_id)

                            if meas_obj:
                                """PROCESSING"""
                                if np.logical_or(process == 2, process == 40):
                                    logger.info("Start Processing Slot %s" % slot_id)
                                    request = scc_obj.session.get(
                                        meas_obj.rerun_all_url, stream=True
                                    )
                                    if request.status_code != 200:
                                        logger.error(
                                            "Could not rerun pre processing for %s. Status code: %s"
                                            % (slot_id, request.status_code)
                                        )
                                """ DOWNLOADING """
                                logger.info(
                                    "Start Download Processed Slot %s" % slot_id
                                )
                                meas_obj = scc_obj.monitor_processing(slot_id)
                                logger.info("End Download Processed Slot %s" % slot_id)
                            else:
                                logger.error("Measurement Slot %s not found." % slot_id)
                            scc_obj.logout()
                        except TimeoutException:
                            logger.error(
                                "Upload file took longer than %d s" % SCC_INFO["MAX_EXEC_TIME"]
                            )
                        except Exception as e:
                            logger.error(str(e))
                            logger.error("Slot %s not downloaded" % slot_id)
                        # finally:
                        # Turn off timer, Restore handler to previous value
                        # signal.alarm(0)
                    else:
                        logger.info(
                            "Slot %s already downloaded from SCC server" % slot_id
                        )

                    """ PLOT OUTPUT SCC SERVER """
                    if np.logical_or.reduce(
                        (process == 0, process == 1, process == 2, process == 60)
                    ):
                        # not slot_in_local, not slot_in_scc,
                        #                    not scc_output_inlocal,
                        logger.info("Plot Output SCC")
                        plot_scc_output(scc_output_slot_dn, scc_code=scc_id)
                    else:
                        logger.info("Plot Output SCC already done")
                else:
                    logger.warning(
                        "Too Early to process Slot %s: %s"
                        % (slot_id, now_dt.strftime("%Y%m%d_%H%M"))
                    )
                    logger.warning(
                        "Slot %s ends at %s"
                        % (slot_id, slots_end[s].strftime("%Y%m%d_%H%M"))
                    )
        else:
            msg = "No Measurement Files found."
            logger.warning(msg)
            return False, msg

    logger.info("End Process day: %s" % date_str)
    if msg == "":
        status = True
    else:
        status = False
    return status, msg


def process_scc(
    lidar_name: str = "alhambra",
    measuremente_type: str = "RS",
    scc_config_id: int = 729,
    raw_dir: Path | str | None = None,
    scc_dir: Path | str | None = None,
    campaign_cfg_fn: str | Path | None = None,
    **kwargs,
):
    """[summary]

    Parameters
    ----------

    Returns
    -------

    """

    """ Start SCC Processing """
    logger.info("Start SCC")

    """Get Input Arguments"""
    if lidar_name not in LIDAR_INFO["lidars"]:
        raise ValueError(f"Lidar {lidar_name} does not registered in GFATPY.")

    # Measurement Type
    if measuremente_type not in MEASUREMENT_TYPES:
        raise ValueError(
            f"Measurement Type {measuremente_type} does not registered in GFATPY. Exit program."
        )

    if raw_dir is None:
        raw_dir = Path.cwd()
    elif isinstance(raw_dir, str):
        raw_dir = Path(raw_dir)

    if not raw_dir.exists():
        raise FileNotFoundError(f"Raw data directory {raw_dir} does not exist")

    if scc_dir is None:
        scc_dir = Path.cwd()
    elif isinstance(scc_dir, str):
        scc_dir = Path(scc_dir)

    if not scc_dir.exists():
        raise FileNotFoundError(f"SCC data directory {scc_dir} does not exist")

    # Campaign Config
    if campaign_cfg_fn is None:
        # SCC lidar config Necessary if No Campaign Config File is given
        if scc_config_id is not None:
            raise ValueError("Campaign config or SCC config ID must be given")
    else:
        if isinstance(campaign_cfg_fn, str):
            campaign_cfg_fn = Path(campaign_cfg_fn)
        if not campaign_cfg_fn.exists():
            raise FileNotFoundError(f"Campaign config {campaign_cfg_fn} does not exist")

    # Dates
    date_ini_str = kwargs.get("date_ini_str", None)
    if date_ini_str is None:
        date_ini_dt = dt.datetime.utcnow()
    else:
        date_ini_dt = dt.datetime.strptime(date_ini_str, "%Y%m%d")
    date_end_str = kwargs.get("date_end_str", None)
    if date_end_str is None:
        date_end_dt = date_ini_dt
    else:
        date_end_dt = dt.datetime.strptime(date_end_str, "%Y%m%d")

    # Start Hour:
    hour_ini = kwargs.get("hour_ini", None)

    # End Hour:
    hour_end = kwargs.get("hour_end", None)

    # Hour Resolution: If no campaign config file is given and different from default
    hour_resolution = kwargs.get("hour_resolution", None)
    if hour_resolution is None:
        hour_resolution = 1.0

    # Timestamp:
    timestamp = kwargs.get("timestamp", None)
    if timestamp is None:
        timestamp = 0

    # Slot Name Type:
    slot_name_type = kwargs.get("slot_name_type", None)
    if slot_name_type is None:
        slot_name_type = 1

    # Type of Processing
    process = kwargs.get("process", None)
    if process is None:
        process = 0

    # Mode of Processing: Operational/Offline
    mode = kwargs.get("mode", None)
    if mode is None:
        mode = 0
    if mode == 1:
        do_send_email = False
    else:
        do_send_email = True

    # SCC data format directory: [data_dn/LIDAR/SCCxxx] / user-defined
    logger.info("LIDAR: %s" % lidar_name)

    try:
        """Set SCC Campaign Configuration"""
        scc_campaign_config = get_campaign_config(
            campaign_cfg_fn=campaign_cfg_fn,
            scc_config_id=scc_config_id,
            hour_ini=hour_ini,
            hour_end=hour_end,
            hour_resolution=hour_resolution,
            timestamp=timestamp,
            slot_name_type=slot_name_type,
        )

        """ Process SCC workflow: loop along days """
        if date_end_dt.date() > date_ini_dt.date():
            logger.info("process period: %s - %s" % (date_ini_str, date_end_str))
            date_range = pd.date_range(date_ini_dt, date_end_dt)
            for i_date in date_range:
                i_date_str = i_date.strftime("%Y%m%d")
                result, msg_day = process_day(
                    lidar_name,
                    i_date_str,
                    meas_type=measuremente_type,
                    scc_campaign_config=scc_campaign_config,
                    process=process,
                    mode=mode,
                    raw_dir=raw_dir,
                    scc_dir=scc_dir,
                )
                if not result:
                    msg = "ERROR: Lidar %s. Day %s. %s " % (
                        lidar_name,
                        i_date_str,
                        msg_day,
                    )
                    if do_send_email:
                        send_email(email_sender, email_receiver, msg)
        elif date_end_dt.date() == date_ini_dt.date():
            logger.info("process single day")
            date_str = date_ini_dt.strftime("%Y%m%d")
            result, msg_day = process_day(
                lidar_name,
                date_str,
                meas_type=measuremente_type,
                scc_campaign_config=scc_campaign_config,
                process=process,
                mode=mode,
                raw_dir=raw_dir,
                scc_dir=scc_dir,
            )
            if not result:
                msg = "ERROR: Lidar %s. Day %s. %s " % (lidar_name, date_str, msg_day)
                if do_send_email:
                    send_email(email_sender, email_receiver, msg)
        else:
            msg = "Error in input dates. Exit program"
            logger.error(msg)
            if do_send_email:
                send_email(email_sender, email_receiver, msg)
            return
    except Exception as e:
        logger.error("%s. Exit program" % str(e))
        msg = "Exception Error in process_scc"
        logger.error(msg)
        if do_send_email:
            send_email(email_sender, email_receiver, msg)
        return


def parse_args():
    """Parse Input Arguments
    python -u scc.py -arg1 v1 -arg2 v2 ...

    Parameters
    ----------
    ()

    Returns
    -------
    args: dict
        Dictionary 'arg':value for input

    """
    # TODO: incluir input sobre resolucion de slot y timestamp

    logger.info("Parse Input")

    parser = argparse.ArgumentParser(description="usage %prog [arguments]")

    parser.add_argument(
        "-i",
        "--initial_date",
        action="store",
        dest="date_ini_str",
        help="Initial date [example: '20190131'].",
    )
    parser.add_argument(
        "-e",
        "--final_date",
        action="store",
        dest="date_end_str",  # required=True,
        help="Final date [example: '20190131'].",
    )
    parser.add_argument(
        "-l",
        "--lidar_name",
        action="store",
        dest="lidar_name",
        default="MULHACEN",
        help="Name of lidar system ['MULHACEN', 'VELETA']. Default: 'MULHACEN'.",
    )
    parser.add_argument(
        "-t",
        "--measurement_type",
        action="store",
        dest="meas_type",
        default="RS",
        help="Type of measurement [example: 'RS', 'HF'].",
    )
    parser.add_argument(
        "-c",
        "--campaign_cfg_fn",
        action="store",
        dest="campaign_cfg_fn",
        default="GFATserver",
        help="campaign config file in JSON format (full path), including name and scc lidar configurations.\
              Default: GFATserver means ACTRIS standardized format. \
              File must include the following fields: \
              name: str",
    )
    parser.add_argument(
        "-s",
        "--scc_lidar_cfg",
        type=int,
        action="store",
        dest="scc_lidar_cfg",
        help="SCC lidar configuration [example: 436, 403]\
              IF NO CAMPAIGN CONFIG FILE IS GIVEN.",
    )
    parser.add_argument(
        "-hi",
        "--hour_ini",
        type=float,
        action="store",
        dest="hour_ini",
        help="Start Hour (HH.H) for creation of slots.\
              DEFAULT: First Time of Measurement within the day",
    )
    parser.add_argument(
        "-he",
        "--hour_end",
        type=float,
        action="store",
        dest="hour_end",
        help="End Hour (HH.H) for creation of slots.\
              DEFAULT: Last Time of Measurement within the day",
    )
    parser.add_argument(
        "-r",
        "--hour_resolution",
        action="store",
        type=float,
        dest="hour_resolution",
        default=1,
        help="time resolution measurement slot. Lidar measurements will be \
              splitted in time slots in number of hours (default: 1 hour) \
              IF NO CAMPAIGN CONFIG FILE IS GIVEN.",
    )
    parser.add_argument(
        "-ts",
        "--timestamp",
        type=int,
        action="store",
        dest="timestamp",
        help="Timestamp for slot: 0, beginning of interval; 1: center of interval.\
              DEFAULT: 0",
    )
    parser.add_argument(
        "-snt",
        "--slot_name_type",
        type=int,
        action="store",
        dest="slot_name_type",
        help="Type of slot naming:\
                0: Earlinet:  YYYYMMDD+station+slot_number.\
                1: SCC Campaigns: YYYYMMDD+station+HHMM(Timestamp).\
                2: Alternative: YYYYMMDDHHMM(Timestamp)+station+scc.\
                DEFAULT: 1",
    )
    parser.add_argument(
        "-p",
        "--process",
        action="store",
        type=int,
        dest="process",
        default=0,
        help="what steps to process [-1: whole chain skipping steps if possible, \
                                0 (DEFAULT): whole chain forcing to process all steps, \
                                1: skip 0a->scc format, plot input. do: upload, process, download, plot output\
                                2: skip 0a->scc, plot input, upload to scc. do: process, download, plot_output\
                               10: only 0a->scc format, \
                               20: only plot input scc format, \
                               30: only upload&process to scc, \
                               40: only process&download from scc, \
                               50: only download from scc, \
                               60: only plot downloaded from scc",
    )
    parser.add_argument(
        "-m",
        "--mode",
        action="store",
        type=int,
        dest="mode",
        default=0,
        help="[0: operational (real-time. checks if the slot period has ended before processing it), 1: off-line",
    )
    parser.add_argument(
        "-d",
        "--data_dn",
        action="store",
        dest="data_dn",
        default="GFATserver",
        help="disk directory for all data",
    )
    parser.add_argument(
        "-o",
        "--scc_dn",
        action="store",
        dest="scc_dn",
        default="GFATserver",
        help="directory where scc files transformed from 0a are saved. \
              Default: 'GFATserver' means copy files to the GFAT NAS.",
    )

    args = parser.parse_args()

    return args.__dict__


def main():
    """
    To be called from terminal / shell:
    python -u scc.py -arg1 v1 -arg2 v2 ...

    """

    logger.info("Start SCC: %s" % dt.datetime.utcnow().strftime("%Y%m%d_%H%M"))

    """ parse args """
    args = parse_args()

    """ process scc workflow """
    process_scc(**args)

    logger.info("End SCC: %s" % dt.datetime.utcnow().strftime("%Y%m%d_%H%M"))


if __name__ == "__main__":
    """main"""
    main()
